
MOVE_TO = 0
LINE_TO = 1
CURVE_TO = 2
# these are not actually saved
HLINE_TO = 3
VLINE_TO = 4
SMOOTHCURVE_TO = 5
CLOSE = 6

class Component:
	def __init__(self, type, point1=None, point2=None, point3=None):
		self.type = type
		self.point1 = point1
		self.point2 = point2
		self.point3 = point3
	def dump(self):
		if self.type == MOVE_TO:
			print("Move To", self.point1)
		elif self.type == LINE_TO:
			print("Line To", self.point1)
		elif self.type == CURVE_TO:
			print("Curve To", self.point1, self.point2, self.point3)
		elif self.type == CLOSE:
			print("Close")

def chomp(bpath):
	while bpath and bpath[0] in ' \t\n\r,':
		bpath = bpath[1:]
	return bpath

def read_num(str):
	num = ''
	while str and str[0] in '0123456789.+-':
		num = num + str[0]
		str = str[1:]
	return float(num), chomp(str)

def parse_bpath(bpath):
	p = []

	last_open = (0, 0)
	last_point = (0, 0)
	last_control = (0,0)
	last_type = MOVE_TO
	last_relative = 0

	bpath = chomp(bpath)
	while bpath:
		# grok new commands
		print(bpath)
		if bpath[0] == 'M':
			bpath = chomp(bpath[1:])
			last_type = MOVE_TO
			last_relative = 0
		elif bpath[0] == 'm':
			bpath = chomp(bpath[1:])
			last_type = MOVE_TO
			last_relative = 1
		elif bpath[0] == 'L':
			bpath = chomp(bpath[1:])
			last_type = LINE_TO
			last_relative = 0
		elif bpath[0] == 'l':
			bpath = chomp(bpath[1:])
			last_type = LINE_TO
			last_relative = 1
		elif bpath[0] == 'H':
			bpath = chomp(bpath[1:])
			last_type = HLINE_TO
			last_relative = 0
		elif bpath[0] == 'h':
			bpath = chomp(bpath[1:])
			last_type = HLINE_TO
			last_relative = 1
		elif bpath[0] == 'V':
			bpath = chomp(bpath[1:])
			last_type = VLINE_TO
			last_relative = 0
		elif bpath[0] == 'v':
			bpath = chomp(bpath[1:])
			last_type = VLINE_TO
			last_relative = 1
		elif bpath[0] == 'C':
			bpath = chomp(bpath[1:])
			last_type = CURVE_TO
			last_relative = 0
			last_smooth = 0
		elif bpath[0] == 'c':
			bpath = chomp(bpath[1:])
			last_type = CURVE_TO
			last_relative = 1
			last_smooth = 0
		elif bpath[0] == 'S':
			bpath = chomp(bpath[1:])
			last_type = SMOOTHCURVE_TO
			last_relative = 0
		elif bpath[0] == 's':
			bpath = chomp(bpath[1:])
			last_type = SMOOTHCURVE_TO
			last_relative = 1
		elif bpath[0] in 'Zz':
			bpath = chomp(bpath[1:])
			last_type = CLOSE
			last_relative = 0
		elif bpath[0] not in '0123456789.+-':
			raise TypeError("unexpected input")

		if last_type == MOVE_TO:
			x, bpath = read_num(bpath)
			y, bpath = read_num(bpath)
			if last_relative:
				x = x + last_point[0]
				y = y + last_point[1]
			last_point = (x, y)
			last_control = last_point
			last_open = last_point
			p.append(Component(MOVE_TO, last_point))
		elif last_type == LINE_TO:
			x, bpath = read_num(bpath)
			y, bpath = read_num(bpath)
			if last_relative:
				x = x + last_point[0]
				y = y + last_point[1]
			last_point = (x, y)
			last_control = last_point
			p.append(Component(LINE_TO, last_point))
		elif last_type == HLINE_TO:
			x, bpath = read_num(bpath)
			if last_relative:
				x = x + last_point[0]
			last_point = (x, last_point[1])
			last_control = last_point
			p.append(Component(LINE_TO, last_point))
		elif last_type == VLINE_TO:
			y, bpath = read_num(bpath)
			if last_relative:
				y = y + last_point[1]
			last_point = (last_point[0], y)
			last_control = last_point
			p.append(Component(LINE_TO, last_point))
		elif last_type == CURVE_TO:
			x1, bpath = read_num(bpath)
			y1, bpath = read_num(bpath)
			x2, bpath = read_num(bpath)
			y2, bpath = read_num(bpath)
			x3, bpath = read_num(bpath)
			y3, bpath = read_num(bpath)
			if last_relative:
				x1 = x1 + last_point[0]
				y1 = y1 + last_point[1]
				x2 = x2 + last_point[0]
				y2 = y2 + last_point[1]
				x3 = x3 + last_point[0]
				y3 = y3 + last_point[1]
			last_point = (x3, y3)
			last_control = (x2, y2)
			p.append(Component(CURVE_TO, (x1,y1),(x2,y2),(x3,y3)))
		elif last_type == SMOOTHCURVE_TO:
			x2, bpath = read_num(bpath)
			y2, bpath = read_num(bpath)
			x3, bpath = read_num(bpath)
			y3, bpath = read_num(bpath)
			if last_relative:
				x2 = x2 + last_point[0]
				y2 = y2 + last_point[1]
				x3 = x3 + last_point[0]
				y3 = y3 + last_point[1]
			x1 = 2 * last_point[0] - last_control[0]
			y1 = 2 * last_point[1] - last_control[1]
			last_point = (x3, y3)
			last_control = (x2, y2)
			p.append(Component(CURVE_TO, (x1,y1),(x2,y2),(x3,y3)))
		elif last_type == CLOSE:
			if last_point != last_open:
				p.append(Component(LINE_TO, last_open))
			last_point = last_open
			last_control = last_point
			p.append(Component(CLOSE))
		bpath = chomp(bpath)
	return p

def print_bpath(bpath):
	p = parse_bpath(bpath)
	list(map(Component.dump, p))

if __name__ == '__main__':
	import sys
	if len(sys.argv) > 1:
		print_bpath(sys.argv[1])



# Generates a weighted dependencies graph from DLLs
# Copyright (c) 2001, 2005, 2007 Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, os, re, math, time

g_maxWeight = 1
# a very limited demangling (leaves junk for some forms)
#rDemangle = re.compile("\?(?:\?0)*([^@]+)(?:@@[0ABGHIKPQUVXYZ?1]+|@Z|@)*([^@]*)")
# somewhat better:
# 1) every mangled name starts with '?'
# 2) ctor has '?1' following, dtor has '?0'  and the class@namespaces@@parameters
# 3) other members have '' following function@class_and_namespaces@@parameters

# this is overly simplified: somehow the key appears to have more than one meaning
Operators = { "2" : " new", "3" : " delete",
			  "4" : "=", "5" : ">>", "6" : "<<", "7" : "!", "8" : "==", "9" : "!=",
			  "C" : "->", "Z" : "-", "F" : "--", "G" : "-", "H" : "+", "D" : "*", "E" : "++",
			  "M" : "<", "Y" : "+=", "A" : "[]", "B" : " char const *", "K" : "/" }
rDemangle = re.compile ("\?(?P<tor>\?[01" + "".join(list(Operators.keys())) + "])*(?P<sym>[\w@_]+?(?=@))@@")
# still NOT handling
'''
d:\Projects>undname ?ReleaseAccess@?$CWriteAccess@VCDocEx@app@@@utl@@UAEXXZ
Microsoft (R) C++ Name Undecorator
Copyright (C) Microsoft Corporation 1981-2001. All rights reserved.

Undecoration of :- "?ReleaseAccess@?$CWriteAccess@VCDocEx@app@@@utl@@UAEXXZ"
is :- "public: virtual void __thiscall utl::CWriteAccess<class app::CDocEx>::ReleaseAccess(void)"

Undecoration of :- "??_7CDocMgrView@app@@6B@"
is :- "const app::CDocMgrView::`vftable'"

Undecoration of :- "??_0CUnit@utl@@QAEAAV01@ABV01@@Z"
is :- "public: class utl::CUnit & __thiscall utl::CUnit::operator/=(class utl::CUnit const &)"
'''
# a list of dlls to ignore dependencies of
g_DontFollow = []

def Demangle (symbols) :
	demangled = []
	for s in symbols :
		m = rDemangle.match (s)
		if m :
			#print m.group("tor"), "::", m.group("sym")
			dm = ""
			names = m.group("sym").split ("@")
			# we must not append the .reverse() in the line above otherwise names would become None. WTF? Inplace operation?
			names.reverse()
			if m.group("tor") == "?0" : # constructor
				dm = "::".join(names) + "::" + names[-1]
			elif  m.group("tor") == "?1" : # constructor
				dm = "::".join(names) + "::~" + names[-1]
			elif  m.group("tor") != None and m.group("tor")[0] == "?" and m.group("tor")[1] in list(Operators.keys()) : # constructor
				dm = "::".join(names) + "::operator" + Operators[m.group("tor")[1]]
			else :
				dm = "::".join(names)
			#print " => ", dm
			demangled.append (dm)
		else :
			#print " <unmatched>: ", s
			demangled.append (s)
	return demangled

class Node :
	def __init__ (self, name, depth) :
		self.name = name
		self.deps = {}
		self.depth = depth # where we found it
		# symbol -> use-count
		self.exports = {}
	def AddEdge (self, name, symbols, delayLoad) :
		self.deps[name] = Edge(name, symbols, delayLoad)
	def AddExports (self, symbols, used=0) :
		''' remember the symbols this node exports (and their use)'''
		for s in symbols :
			if s in self.exports :
				self.exports[s] += used
			else :
				self.exports[s] = used

class Edge :
	def __init__ (self, name, symbols, delayLoad) :
		global g_maxWeight
		self.name = name # the target
		self.reduce = 0 # != 0 if this should better vanish
		self.delayLoad = delayLoad
		if len(symbols) > g_maxWeight :
			g_maxWeight = len(symbols)
		self.symbols = Demangle(symbols)
	def Weight (self) :
		"As method to have let always be the number of symbols"
		return len(self.symbols)
g_path = None # empty, default behaviour
def FindInPath (sName) :
	"returns the complete path of the given name, looking at cwd first"
	sDir = os.getcwd()
	p = sDir + os.sep + sName
	if os.path.exists (p) :
		return p
	myPath = os.environ['PATH']
	if g_path :
		myPath = g_path
	for sDir in myPath.split (os.pathsep) :
		p = sDir + os.sep + sName
		if os.path.exists (p) :
			return p
	# safety
	return sName

def GetDepsDotnet (sFrom, dAll, node, nMaxDepth, nDepth=0) :
	""" passing in a node allows to check C++ and .Net dependecies """
	try :
		import clr
		clr.AddReference("IronPython")
	except ImportError :
		print("No Iron!")
		return
	from System.Reflection import Assembly

	if sFrom not in dAll or node :
		if not node :
			node = Node (sFrom, nDepth)
			dAll[sFrom] = node
		try :
			# assembly = Assembly.LoadFrom(sFrom)
			# the former does not show C++ -> .Net connections
			assembly = Assembly.ReflectionOnlyLoadFrom(sFrom)
		except IOError :
			return
		except SystemError :
			return
		symbols = []

		try :
			types = assembly.GetExportedTypes()
		except IOError :
			# some failed dependencies are showing up here?
			return
		for t in types :
			#if not t.IsPublic :
			#	continue
			if t.IsClass or t.IsInterface :
				members = t.GetMembers()
				for  m in members :
					if m.Module.Assembly == assembly :
						#print "Exp:", sym
						pass
					elif t.IsImport :
						#print "Imp:", t.Name + "." + m.Name + "()"
						sym = t.Name + "." + m.Name + "()"
						symbols.append (sym)
		# we may want to look at all refernced assemblies ...
		refs = assembly.GetReferencedAssemblies()
		for r in refs :
			n2 = r.Name + ".dll";
			print("\t" * nDepth + n2, r.Version)
			symbols = [ "[" + str(r.Version) + "]" ]
			node.AddEdge (n2, symbols, 1) # delayLoad: kind of
			GetDepsDotnet (n2, dAll, None, nMaxDepth, nDepth+1)

def GetDepsWin32 (sFrom, dAll, nMaxDepth, nDepth=0) :
	"calculates the dependents of the passed in dll"
	if nMaxDepth <= nDepth :
		return
	sFrom = sFrom.lower()
	global g_DontFollow
	if sFrom in g_DontFollow :
		dAll[sFrom] = Node (sFrom, nDepth) # needed here so other algoritm can remove it ;)
		return
	if sFrom not in dAll :
		node = Node (sFrom, nDepth)
		sPath = sFrom
		if g_path :
			sPath = FindInPath (sFrom)

		# first remember all exports
		arr = []
		f = os.popen ('dumpbin /exports "' + sPath + '"')
		sDump = f.readlines()
		for s in sDump :
			# similiar regex as above, but we have three numbers
			r2 = re.match ("^[ ]+(?:[0123456789ABCDEF]{1,5}[ ]+)(?:[0123456789ABCDEF]{1,5}[ ]+)?[0123456789ABCDEF]{8}[ ]+([\w@?$]+).*", s)
			if r2 :
				arr.append (r2.group(1))
		node.AddExports (arr, 0)

		f = os.popen ('dumpbin /imports "' + sPath + '"')
		name = None
		arr = []
		sDump = f.readlines ()
		# avoids multiple instances of dumpbin running simultaneously
		f = None
		# avoids infinite recursion on circular dependencies
		dAll[sFrom] = None
		delayLoad = 0
		directDeps = []
		for s in sDump :
			r = re.match ("^    (.*\.dll)", s, re.IGNORECASE)
			# the delay load switch is flipped once
			if s.find ("delay load imports") > 0 :
				delayLoad = 1
			if r :
				name = r.group(1).lower()
				# print name
			else :
				# import by name
				# The system dlls seem to follow a diferent pattern (or is this for know-dlls? Delay load?)  thus (?:[0123456789ABCDEF]{8}[ ]+)?
				r2 = re.match ("^[ ]+(?:[\dA-F]{8}[ ]+)?[\dA-F]{1,5}[ ]+([\w@?$]+)$", s)
				if not r2 :
					r2 = re.match ("^[ ]+(?:[\dA-F]{8}[ ]+)?Ordinal[ ]+([\d]+)$", s)
				if r2 :
					arr.append (r2.group(1))
				elif s[:-1] == "" and name != None and len(arr) > 0 :
					#print name, len(arr)
					node.AddEdge (name, arr, delayLoad)
					arr = []
					directDeps.append(name)
					GetDepsWin32 (name, dAll, nMaxDepth-nDepth+2, nDepth+1)
		# add to all nodes
		dAll[sFrom] = node

		# check for dotnet dependencies
		if "mscoree.dll" in directDeps :
			print("Dotnet check ...", sFrom)
			GetDepsDotnet (sFrom, dAll, node, nMaxDepth-nDepth+2, nDepth+1)

		# restore original depth (independent of how the recursion works)
		for sd in directDeps :
			if sd in list(dAll.keys()) :
				try :
					dAll[sd].depth = nDepth + 1
				except AttributeError as msg :
					print("FIXME:", sd, msg)

def GetDepsPosix (sFrom, dAll, nMaxDepth, nDepth=0) :
	"calculates the dependents of the passed in so"
	# sFrom must be with absolute on Posix, still we prefer the shorter name
	sFromName = os.path.split(sFrom)[-1]
	if nMaxDepth <= nDepth :
		return
	global g_DontFollow
	if sFromName in g_DontFollow :
		dAll[sFromName] = Node (sFromName, nDepth) # needed here so other algoritm can remove it ;)
		return
	if not sFromName in list(dAll.keys()) :
		print("Creating", sFromName, nDepth)
		node = Node (sFromName, nDepth)
		sPath = sFrom
		#TODO: work with relative pathes? Current dir?
		if os.sys.platform in ['darwin'] :
			f = os.popen ('dyldinfo -dylibs "' + sPath + '"')
		else :
			f = os.popen ('ldd "' + sPath + '"')
		sModules = f.readlines ()
		# avoids multiple instances of dumpbin running simultaneously
		f = None
		# avoids infinite recursion on circular dependencies
		dAll[sFromName] = None
		# sFrom imports
		# --extern-only : -g
		# --undefined-only : -u
		# ? --defined-only : -U
		if os.sys.platform in ['darwin'] :
			fImports = os.popen ('nm -g -u ' + sPath)
		else :
			fImports = os.popen ('nm --extern-only --dynamic --undefined-only ' + sPath)
		sImports = fImports.readlines()
		dImports = {}
		for si in sImports :
			if os.sys.platform in ['darwin'] :
				if len(si) > 1 :
					dImports[si[:-1]] = 1
			else :
				mi = re.match("\s+U (\S+)", si)
				if mi :
					print(mi.group(1))
					dImports[mi.group(1)] = 1
		toRecurse = []
		for s in sModules :
			if os.sys.platform in ['darwin'] :
				r = re.match ("\s+(?P<path>\S+).*", s)
			else :
				r = re.match ("\s+(?P<name>\S+) => (?P<path>\S+).*", s)
			if r :
				# print r.group("name"), r.group("path")
				# now find symbols between sFrom and any of the SOs
				usedSymbols = []
				unusedSymbols = []
				if os.sys.platform in ['darwin'] :
					fExports = os.popen ('nm -g -U ' + r.group("path"))
				else :
					fExports = os.popen ('nm --extern-only --dynamic --defined-only ' + r.group("path"))
				sExports = fExports.readlines()
				for se in sExports :
					me = re.match ("\w+ T (\S+)", se)
					if me :
						symbol = me.group(1)
						# print symbol
						# print me.group(1), r.group('name')
						if symbol in list(dImports.keys()) :
							# direct connection
							usedSymbols.append (symbol)
						else :
							unusedSymbols.append(symbol)
				# now we have complete symbols for node r.group('name'),
				# should remember the node although it may not be connected yet
				# OTOH the recursion would become more difficult to understand, favor KISS
				if len(usedSymbols) > 0 :
					try :
						node.AddEdge(r.group("name"), usedSymbols, 0)
					except IndexError : # darwin
						sToPath = r.group("path")
						node.AddEdge(sToPath[sToPath.rfind('/')+1:], usedSymbols, 0)
					toRecurse.append(r.group("path"))
					print(sFromName, "(", nDepth, "):", r.group("path"))
		# add to all nodes
		dAll[sFromName] = node
		for sd in toRecurse :
			print(sFromName, "->", sd, "level:", nDepth+1)
			sdn = os.path.split(sd)[-1]
			GetDepsPosix (sd, dAll, nMaxDepth-nDepth+2, nDepth+1)
			if sdn in list(dAll.keys()) :
				#print "Adjusting depth", sd, nDepth+1
				dAll[sdn].depth = nDepth+1 # adjust to original depth (not how our recursion found it)

def IsWin32 () :
	return os.name in ["nt"]

def Remove (deps, list) :
	"From the deps tree remove the nodes matching list"
	for s in list :
		if s in deps :
			del deps[s]
		for sn in list(deps.keys()) :
			node = deps[sn]
			if s in node.deps :
				del node.deps[s]
def RemoveRegEx (deps, reg_ex) :
	"If a module matches reg_ex remove it"
	import re
	rr = re.compile (reg_ex)
	for k1 in list(deps.keys()) :
		if rr.match (k1) :
			del deps[k1]
		else :
			node = deps[k1]
			for k2 in list(node.deps.keys ()) :
				if rr.match (k2) :
					del node.deps[k2]
def RemoveNonLocal (deps) :
	"Remove everything not available in the current directory"
	for k in list(deps.keys()) :
		node = deps[k]
		for c in list(node.deps.keys()) :
			if not os.path.exists (c) :
				del node.deps[c]
	# also remove from the root
	root_keys = list(deps.keys())
	for k in root_keys :
		if not os.path.exists (k) :
			del deps[k]

def RemoveBySymbols (deps, list) :
	"If a connection is conly caused by some symbol in 'list' it is removed"
	for k in list(deps.keys()) :
		node = deps[k]
		for c in list(node.deps.keys ()) :
			edge = node.deps[c]
			kills = []
			for s2 in edge.symbols :
				for s1 in list :
					# robustness, dont compare the empty string
					if s1 == "" : continue
					# only comparing the start of the symbol
					if s2.find (s1) == 0 :
						#print "removing", s2, "from", c, "->", k
						kills.append (s2)
						#NOT modifying while iterating: edge.symbols.remove (s2)
						break
			if len(edge.symbols) == len(kills) :
				# remove complete edge (maybe we should only mark it removed?)
				#print "removing", c, "from", k
				del node.deps[c]
			else :
				for s in kills :
					edge.symbols.remove (s)
		# remove the symbols also from the dependency edge is pointing to
		for s in list :
			if s in list(node.exports.keys()) :
				del node.exports[s]

def Reduce (deps, f, bHintOnly = 1) :
	"Automatically remove connections until there is only something reasonable left"
	# first iteration: two components are connected in both directions
	# if one connection is much weaker than the other one, the weaker one is considered bad
	# if both have similar weight they are both treated as bad and removed
	keys = list(deps.keys())
	keys.sort()
	nReduced = 0
	for k in keys :
		node = deps[k]
		for c in list(node.deps.keys ()) :
			edge = node.deps[c]
			if c in deps :
				node2 = deps[c]
				if k in node2.deps :
					edge2 = node2.deps[k]
					# if one already is reduced dont do it again
					if edge.reduce or edge2.reduce :
						continue
					n = len(edge.symbols)
					n2 = len(edge2.symbols)
					if  abs(n - n2) < 3 :
						f.write ("Remove (" + str(n) + "," + str(n2) +  ") " + edge.name + " <-> " + edge2.name +
							"\n\t" + ", ".join (edge.symbols) +
							"\n\t" + ", ".join (edge2.symbols) + "\n")
						# if we now remove *both* connections some components could become completely disconnected.
						# Instead may try some more guessing. Maybe the component with less deps should drop one more ?
						if bHintOnly :
							edge.reduce = 1
							edge2.reduce = 1
						else :
							if c in node.deps : del node.deps[c]
							if k in node2.deps : del node2.deps[k]
						nReduced += (n + n2)
					elif n > n2 :
						f.write ("Remove (" + str(n2) + ") " + edge.name + " -> " + edge2.name +
							"\n\t" + ", ".join (edge2.symbols) + "\n")
						if bHintOnly :
							edge2.reduce = 1
						else :
							if k in node2.deps : del node2.deps[k]
						nReduced += n2
					else :
						f.write("Remove (" + str(n) + ") " + edge2.name + " -> " + edge.name +
							"\n\t" + ", ".join (edge.symbols) + "\n")
						if bHintOnly :
							edge.reduce = 1
						else :
							if c in node.deps : del node.deps[c]
						nReduced += n
	f.write("Remove total: %d\n" % (nReduced))

def CutLeafs (deps, nCuts) :
	"Remove components without connections. After some iterations only circulars remain (or nothing)"
	leafs = {}
	cut = [] # what we have cut in the last pass
	while nCuts > 0 :
		cut = list(leafs.keys())
		leafs = {}
		# first pass: find leafs
		for k in list(deps.keys()) :
			node = deps[k]
			if len(list(node.deps.keys ())) == 0 :
				leafs[k] = 0
		if len(list(leafs.keys())) == 0 :
			break # nothing remaining
		nCuts -= 1
		# second pass : remove references to leafs
		for k in list(deps.keys()) :
			node = deps[k]
			for c in list(leafs.keys()) :
				if c in node.deps :
					leafs[c] += 1
					del node.deps[c]
		# third pass: remove the leafs
		for k in list(leafs.keys()) :
			# we either need to reset leafs in every round or
			if k in deps :
				del deps[k]
	if len(list(leafs.keys())) > 0 :
		return list(leafs.keys())
	return cut
def TransitiveReduction (deps, tops, treds=[]) :
	"""If a component A has B and C as dependency and B also depends on C reduce A->C.
	   This algorithm is available with dot, too. But here we can adjust the respective
	   'inherited' symbols and 'depth' as well.
	"""
	for a in tops :
		bs = list(deps[a].deps.keys())
		cs = {}
		for b in bs :
			for c in list(deps[b].deps.keys()) :
				if not c in list(cs.keys()) :
					cs[c] = 1
		# bs and cs are now complete, remove all a->b, where b in cs
		extra_syms = {}
		for c in list(cs.keys()) :
			if c in bs :
				print("Cut", a, "->", c)
				# we need to know which edge a->b is taking the extra symbols
				extra_syms[c] = deps[a].deps[c].symbols
				del deps[a].deps[c]
		# fill remaining bs having a c dependency with the extra symbols
		e_keys = list(extra_syms.keys())
		for e in e_keys :
			bs = list(deps[a].deps.keys()) # they are reduced above
			for b in bs :
				if e in list(deps[b].deps.keys ()) :
					print("Adjust", a, "->", b, "+", len(extra_syms[e]), "from", e)
					#NOT: deps[a].deps[b].symbols.extend(extra_syms[e]) # producing duplicates
					for s in extra_syms[e] :
						if not s in deps[a].deps[b].symbols :
							deps[a].deps[b].symbols.append(s)
					del extra_syms[e]
					break
		#FIXME: something we have cut before may not be visible anymore (would need a deeper search or an ordered reduction)
		if len(list(extra_syms.keys())) :
			print("Not adjusted:", a, "->", ", ".join(list(extra_syms.keys())))
	# recurse into remaining bs
	for a in tops :
		bs = list(deps[a].deps.keys())
		treds.append(a) # rember already done to avoid endless recursion
		TransitiveReduction (deps, bs, treds)

def TopMost (deps) :
	"everything not used by something else"
	keys = list(deps.keys ())
	topmost = []
	used = {}
	for k in keys :
		for d in list(deps[k].deps.keys()) :
			if d not in used :
				used[d] = 1
	for k in keys :
		if k not in used :
			topmost.append (k)
	return topmost

def RecalculateDepth (deps) :
	"From the top-most - i.e. unused - down to the bottom level "
	keys = list(deps.keys ())
	used = {}
	tops = []
	depth = 0
	while 1 :
		top = keys
		for k in keys :
			for d in list(deps[k].deps.keys()) :
				used[d] = 1
		for k in keys :
			if not k in list(used.keys ()) :
				tops.append (k)
		if len(tops) == 0 :
			break
		for k in tops :
			deps[k].depth = depth
		keys = [x for x in keys if not x in tops]
		tops = []
		depth += 1

def CalculateUsed (deps) :
	"Given the complete dependency tree calcualte the use count of every symbol"
	for sn in list(deps.keys()) :
		node = deps[sn]
		edge_keys = list(node.deps.keys())
		for se in edge_keys :
			edge = node.deps[se]
			target = deps[se]
			# every edges symbols
			target.AddExports(edge.symbols, 1)

def UnGlob (comps) :
	import glob
	comp_dict = {}
	for p in comps :
		g = glob.glob (p)
		for d in g :
			if d not in comp_dict :
				comp_dict[d] = 1
	return list(comp_dict.keys())

def Sorted (dict) :
	"given a dictionary with name to number, sort by number"
	ret = []
	keys = list(dict.keys())
	keys.sort ( lambda a, b : cmp(dict[a], dict[b]) )
	for k in keys :
		ret.append ((k, dict[k]))
	return ret
# some predefined sets of DLLs, either for hiding from the dependencies or maybe to tin them
dllsSysWin32 = [
	"version.dll", "winmm.dll",
	"kernel32.dll", "user32.dll", "gdi32.dll", "comdlg32.dll", "advapi32.dll", "shell32.dll",
	"comctl32.dll", "ole32.dll", "oleaut32.dll", "winspool.drv", "imm32.dll",
	"ws2_32.dll", "ws2help.dll", "wsock32.dll", "ntdll.dll", "mpr.dll",
	"rpcrt4.dll", "shlwapi.dll", "netapi32.dll", "msimg32.dll", "oledlg.dll",
	"setupapi.dll", "secur32.dll", "avifil32.dll", "msvfw32.dll",
	"dbghelp.dll", "uxtheme.dll", "dnsapi.dll", "activeds.dll", "crypt32.dll",
	"mscoree.dll", "psapi.dll", "msi.dll", "mscms.dll", "glu32.dll", "hid.dll", "imagehlp.dll",
	"powrprof.dll", "quartz.dll", "wininet.dll", "wintrust.dll", "riched20.dll" , "odbc32.dll",
	"oleacc.dll", "shfolder.dll", "opengl32.dll"
	]
dllsCrts = [
	"msvcrt.dll", "msvcrtd.dll", "msvcp60.dll",
	"msvcrt20.dll",
#	"msvcr70.dll", "msvcp70.dll",
	"msvcr70.dll", "msvcr70d.dll", "msvcp70.dll", "msvcp70d.dll",
	"msvcr71.dll", "msvcr71d.dll", "msvcp71.dll", "msvcp71d.dll",
	"msvcr80.dll", "msvcr80d.dll", "msvcp80.dll", "msvcp80d.dll",
	"msvcr90.dll", "msvcr90d.dll", "msvcp90.dll", "msvcp90d.dll", "vcomp90.dll", "vcomp90d.dll",
	"msvcr100.dll", "msvcr100d.dll", "msvcp100.dll", "msvcp100d.dll", "vcomp100.dll", "vcomp100d.dll",
	# only one on Linux? (or am I missing th C++rt, and what about librt.so.1?)
	"libc.so.5", "libc.so.6"
	]
dllsMfc = [
	"mfc42u.dll", "mfc42.dll", "mfc42ud.dll",
	"mfc70u.dll", "mfc70.dll", "mfc70ud.dll",
	"mfc71u.dll", "mfc71.dll", "mfc71ud.dll",
	"mfc80u.dll", "mfc80.dll", "mfc80ud.dll", "atl80.dll",
	"mfc90u.dll", "mfc90.dll", "mfc90ud.dll",
	"mfc100u.dll", "mfc100.dll", "mfc100ud.dll"
	]
dllsGtk = [
	"libglib-2.0-0.dll", "libgmodule-2.0-0.dll", "libgobject-2.0-0.dll", "libgthread-2.0-0.dll",
	"libpango-1.0-0.dll", "libpangowin32-1.0-0.dll", "libpangoft2-1.0-0.dll", "libpangocairo-1.0-0.dll",
	"libgdk_pixbuf-2.0-0.dll", "libgdk-win32-2.0-0.dll", "libgtk-win32-2.0-0.dll", "libatk-1.0-0.dll",
	"libcairo.dll", "libintl-1.dll", "iconv.dll"
	]

# Tinting themes
colorMatcher = []
tangoColors = [
	"white", # noone is going to ask for it (start of graph)
	"#fce94f", # butter   (light occer)
	"#fcaf3e", # orange
	"#e9b96e", # chocolate (light brown)
	"#8ae234", # chameleon (green)
	"#729fcf", # sky blue
	"#ad7fa8", # plum
	# leaving out red, now getting darker
	"#edd400",
	"#f57900",
	"#c17d11",
	"#73d216",
	"#3465a4",
	"#75507b",
	# even darker
	"#c4a00",
	"#ce5c00",
	"#8f590f",
	"#4ce9a06",
	"#204a87",
	"#5c3566",
]
def ParseRegexTheme (fname) :
	global colorMatcher
	rkv = re.compile(r'^"(?P<regex>[^"]+)"\s*=\s*(?P<color>[\w#]+).*')
	f = open (fname, 'r')
	for s in f.readlines() :
		m = rkv.match(s)
		if m :
			try :
				sr = re.compile(m.group('regex'))
				sc = m.group('color')
				colorMatcher.append ((sr, sc))
			except re.error as msg :
				print("Parser error\n", s)
				sys.exit(2)
	if len(colorMatcher) == 0 :
		print("no colors parsed:", fname)
		sys.exit(3)
	else :
		print(len(colorMatcher), "colors parsed:", fname)

def LookupColor (node) :
	'''Used to tint the nodes by layer '''
	global colorMatcher
	if len(colorMatcher) :
		for rx, color in colorMatcher :
			m = rx.match(node.name)
			if m and m.start() == 0 and m.end() == len(node.name) :
				#print color, node.name
				return color
		return "white"
	# tinting by palette
	try :
		return tangoColors[node.depth]
	except KeyError as msg :
		print(msg)
		return "#ef2929" # scarlet red
	except IndexError as msg :
		print("Dependency", node.name, 'level', node.depth)
		return "#cc0000" # scarlet ret (a bit darker)

def DumpSymbols (deps, f) :
	Symbols = {}
	Modules = {}
	Imports = {}
	node_keys = list(deps.keys())
	node_keys.sort()
	for sn in node_keys :
		node = deps[sn]
		if len(list(node.deps.keys())) == 0 :
			continue
		f.write (sn + "\n")
		Imports[sn] = 0
		edge_keys = list(node.deps.keys())
		edge_keys.sort()
		for se in edge_keys :
			edge = node.deps[se]
			if edge.reduce :
				f.write ("\t!" + node.name + " -> " + edge.name + "\n")
				continue
			if edge.delayLoad :
				f.write ("\t" + node.name + " -> " + edge.name + " (delay load)\n")
			else :
				f.write ("\t" + node.name + " -> " + edge.name + "\n")
			syms = edge.symbols
			syms.sort()
			for sy in syms :
				f.write ("\t\t" + sy + "\n")
				if sy in Symbols :
					Symbols[sy] += 1
				else :
					Symbols[sy] = 1
			Imports[sn] += len(syms)
			if edge.name in Modules :
				Modules[edge.name].append (node.name)
			else :
				Modules[edge.name] = [node.name]
	# symbols used by many modules are good candidates for refactoring
	f.write ( "***** Modules with users (symbols) *****\n" )
	for s, i in Sorted (Imports) :
		if s in Modules :
			f.write (s + " (" + str(i) + ") : " + ",".join (Modules[s]) + "\n")
		else :
			f.write (s + " (0) : <no users>\n")

def DumpUnusedSymbols (deps, f) :
	for sd in list(deps.keys ()) :
		node = deps[sd]
		for se in list(node.deps.keys()) :
			edge = node.deps[se]
			d = deps[se]
			for ss in edge.symbols :
				if ss in list(d.exports.keys()) :
					del d.exports[ss]
	# after we have removed *all* used symbols
	for sd in list(deps.keys ()) :
		node = deps[sd]
		unused = list(node.exports.keys())
		f.write (sd + "\n")
		unused.sort ()
		for ss in unused :
			f.write ("\t" + ss + "\n")

def DumpSymbolsUse (deps, f) :
	f.write ("*** Symbol Use Count ***\n")
	CalculateUsed (deps)
	node_keys = list(deps.keys())
	node_keys.sort()
	for sn in node_keys :
		node = deps[sn]
		used = 0
		for k, v in node.exports.items() :
			if v > 0 :
				used += 1
		f.write (node.name + " (%d:%d)\n" % (used, len(node.exports)))
		sorted_symbols = Sorted(node.exports)
		for sym, cnt in sorted_symbols :
			try :
				f.write ("\t%d\t%s\n" % (cnt, sym))
			except KeyError :
				f.write ("\t?\t" + sym + "\n")

def DumpReverse (deps, f) :
	node_keys = list(deps.keys ())
	edges = {}
	for sn in node_keys :
		node = deps[sn]
		edge_keys = list(node.deps.keys ())
		for se in edge_keys :
			if se in list(edges.keys ()) :
				edges[se].append ((node.deps[se], sn))
			else :
				edges[se] = [(node.deps[se], sn)]
	edge_keys = list(edges.keys ())
	edge_keys.sort ()
	for se in edge_keys :
		f.write (se + "\n")
		for ed in edges[se] :
			f.write ("\t" + se + " <- " + ed[1] + "\n")
			syms = ed[0].symbols
			syms.sort ()
			for sym in syms :
				f.write ("\t" * 2 + sym + "\n")

def RemoveStrays (deps) :
	"Remove every node which has and is no dependency"
	strays = {}
	# does it have no dependencies?
	for k in list(deps.keys()) :
		if len(list(deps[k].deps.keys())) == 0 :
			strays[k] = 1
	# even if it does not have a dependency it still may be a dependency
	stray_keys = list(strays.keys())
	for k in list(deps.keys()) :
		node = deps[k]
		for d in list(node.deps.keys ()) :
			if d in stray_keys :
				if d in list(strays.keys ()) : # delete it only once ;)
					#print "Not stray:", d
					del strays[d]
	stray_keys = list(strays.keys())
	for k in stray_keys :
		print("Stray:", k)
		del deps[k]

def CreateDsm (deps) :
	"Given the complete dependency tree create a dsm sorted by 'leafs'"
	RemoveStrays(deps)
	nTotal = len(list(deps.keys()))
	nDone = 0
	# for DSM row and colum index are the same, this is mapping the module name to it's index (row number)
	table_map = {}
	node_keys = list(deps.keys())
	node_keys.sort()
	level = 0 # the iteration where we cut the leafs
	node_levels = {}
	while nDone < nTotal :
		leafs = []
		strays = []
		for sn in node_keys :
			node = deps[sn]
			nRefs = 0
			for d in list(node.deps.keys()) : # the dependencies
				if d in node_keys : # still referenced
					nRefs = nRefs + 1
			if nRefs == 0 :
				# not removing just now to treat all leafs on this layer the same
				leafs.append(sn)
		if len(leafs) == 0 :
			# no endless loop for circular dependencies
			histo = {} # map number of dependencies to module
			for sn in node_keys :
				node = deps[sn]
				num = len(list(node.deps.keys()))
				if not num in list(histo.keys()) :
					histo[num] = []
				histo[num].append(sn)
			# find modules with lowest number of dependencies
			for num in range(1, len(list(deps.keys()))+1) :
				if num in list(histo.keys()) :
					leafs = histo[num]
					print("Circular cut:", leafs)
					break
		for sn in leafs :
			table_map[sn] = (nTotal - nDone, level)
			nDone = nDone + 1
			node_levels[sn] = level
		level = level + 1
		# everything which has no reference into our tree or is allready processed
		node_keys = [x for x in node_keys if not x in list(table_map.keys())]
	keys = Sorted (table_map)
	# now we know the index of every module in the dsm, fill cells with number of symbols
	dsm = []
	for y, n in keys :
		# a row has everything which depends on o specific module
		target = deps[y]
		row = []
		for x, m in keys :
			# a colum has all the dependencies of a module
			source = deps[x]
			if y in list(source.deps.keys()) :
				edge = source.deps[y]
				# referencing the object, to later have access to all symbols, instead of just it's number
				row.append (edge)
			else :
				# there is no need to specially mark x==y, i.e. the diagonal
				row.append (None)
		dsm.append (row)
	# we no only need the matrix, but also the order of rows and columns
	keys_alone = []
	for k, n in keys :
		keys_alone.append (k)
	return (keys_alone, dsm, node_levels)

def SaveDsm (deps, f) :
	# create and save as csv
	# Yippie ki-yay - Excel csv interpretation is locale dependent - Control Panel/Region/Numbers/List separator
	csv_list_delimiter = ','
	k, m, levels = CreateDsm (deps)
	f.write ("Level;%s%s\n" % (csv_list_delimiter, csv_list_delimiter.join (k)))
	for y in range(0, len(k)) :
		# first column: the level of dependency
		f.write ("%d" % (levels[k[y]]) + csv_list_delimiter)
		f.write (k[y])
		for x in range(0, len(k)) :
			edge = m[y][x]
			if edge :
				f.write ("%s%d" % (csv_list_delimiter, edge.Weight()))
			else :
				f.write (csv_list_delimiter)
		f.write ("\n")

def SaveDB (deps, fname) :
	# create an sqlite database
	import sqlite3

	con = sqlite3.connect (fname)
	cur = con.cursor()

	# create a table containing data for reproduction
	cur.execute ('CREATE TABLE Settings (Key text, Value text)')
	cur.execute ('INSERT INTO Settings VALUES(?, ?)', ("cwd", os.getcwd()))
	cur.execute ('INSERT INTO Settings VALUES(?, ?)', ("args", ' '.join (sys.argv[1:])))
	con.commit()

	# create the dsm to have all data easily available
	k, m, levels = CreateDsm (deps)

	# create and fill the Modules table
	cur.execute ('CREATE TABLE Modules (Name text,' \
		' ImportedModules integer, ExportedSymbols integer, Level integer)')
	for y in range(0, len(k)) :
		# the number of imports - module count
		num_imp = len(list(deps[k[y]].deps.keys()))
		num_exp = len(list(deps[k[y]].exports.keys()))
		cur.execute ('INSERT INTO Modules VALUES(?, ?, ?, ?)', (k[y], num_imp, num_exp, levels[k[y]]))
	con.commit()
	# create an index of ModuleNames
	cur.execute ('CREATE UNIQUE INDEX ModuleNames ON Modules (Name)')
	con.commit()

	# create and fill the Dependencies table
	cur.execute ('CREATE TABLE Dependencies (' \
		'Consumer text, Provider text, Imports integer, Weight real)')
	for y in range(0, len(k)) :
		for x in range(0, len(k)) :
			edge = m[y][x]
			if edge :
				cur.execute ('INSERT INTO Dependencies VALUES(?,?,?,?)',
					(k[x], k[y], len(edge.symbols), edge.Weight()))
	con.commit ()

	# create and fill the table of Symbols
	cur.execute ('CREATE TABLE Symbols (Name text, User text, Provider text)')
	for y in range(0, len(k)) :
		for x in range(0, len(k)) :
			edge = m[y][x]
			if edge :
				for s in edge.symbols :
					cur.execute ('INSERT INTO Symbols VALUES(?,?,?)', (s, k[x], k[y]))
	# add unused symbols, too
	for sm in k :
		node = deps[sm]
		for e, used in node.exports.items() :
			if used == 0 : # the used ones are already added above
				cur.execute ('INSERT INTO Symbols VALUES(?,?,?)', (e, None, sm))
	con.commit ()

	# create a read-only table accumulating Symbols by Users
	# SELECT Provider,max(Users),count(distinct Name) FROM Symbols GROUP BY Provider
	cur.execute ('CREATE VIEW SymbolsByUse AS SELECT ' \
		'Name,Provider,count(distinct User) AS UseCount FROM Symbols GROUP BY Name')
	con.commit ()

	# another view to sort Modules by number of Consumer
	cur.execute ('CREATE VIEW ModulesByUse AS SELECT ' \
		'Provider,count(distinct Consumer),max(Imports),min(Imports) FROM Dependencies GROUP BY Provider')
	con.commit()

	# yet another view to check how much modules symbols are used
	cur.execute ('CREATE VIEW ModuleCoverageTemp AS SELECT ' \
		'Provider,count(distinct Name) AS UsedSymbols FROM Symbols AS S WHERE User>0 GROUP BY Provider')
	# merge it with Modules.ExportedSymbols
	cur.execute ('CREATE View ModuleCoverage AS SELECT ' \
		'Provider,Level,UsedSymbols,ExportedSymbols,100.0*UsedSymbols/ExportedSymbols AS Coverage ' \
		'FROM Modules AS M JOIN ModuleCoverageTemp AS C ON M.Name=C.Provider')
	con.commit()

	# calculate FanIn (UsingModules) and FanOut (ImportedModules)
	# ,1.0-(1.0/(FanIn+FanOut)) AS Complexity -- needs an extra run
	# due to sqlite3.OperationalError: no such column: FanIn
	cur.execute ('CREATE VIEW ModuleCouplingTemp AS SELECT ' \
		'Name,ImportedModules AS FanOut,count(distinct Consumer) AS FanIn ' \
		'FROM Dependencies AS D JOIN Modules AS M ON D.Provider=M.Name ' \
		'GROUP BY Provider')
	cur.execute ('CREATE VIEW ModuleCoupling AS SELECT ' \
		'Name,FanOut,FanIn,1.0-(1.0/(FanIn+FanOut)) AS Coupling ' \
		'FROM ModuleCouplingTemp')
	con.commit()

def CreateList (s) :
	"Return a list of strings either by split or from file"
	lst = []
	if s[0] == "@" :
		f = open (s[1:])
		for line in f.readlines() :
			if not line[0] in ["#", ";"] :
				lst.append (line[:-1].lower())
	else :
		lst = s.split (",")
	return lst

def ImportDump (sfDump, deps) :
	print("Import from:", sfDump)
	global g_DontFollow
	deps = {}
	f = open (sfDump)
	s = f.readline()
	rNode = re.compile ("^([\w_-]+\.\w+)$")
	rEdge = re.compile ("^\t([\w_-]+\.\w+) -> ([\w_-]+\.\w+)$")
	rSym = re.compile ("^\t\t(.*)$")
	curNode = None
	curEdge = None
	symbols = []
	while s :
		s = f.readline()
		m = rNode.match (s)
		if m :
			# store previous results if
			if curNode and curEdge and len(symbols) :
				if not curNode.name in g_DontFollow :
					curNode.AddEdge (curEdge, symbols, 0) # Todo: delay loaded
				curEdge = None
				symbols = []
			k = m.group(0)
			if k in list(deps.keys()) :
				curNode = deps[k]
			else :
				curNode = deps[k] = Node (k, 0) # Todo: depth?
		else :
			m = rEdge.match(s)
			if m :
				if not curNode.name == m.group(1) :
					print("???", s)
					sys.exit(2)
				else :
					if curNode and curEdge and len(symbols) :
						if not curNode.name in g_DontFollow :
							curNode.AddEdge (curEdge, symbols, 0) # Todo: delay loaded
						#print curNode.name, "=>", curEdge
						symbols = []
					curEdge = m.group(2)
					# have to create the node, too
					if not curEdge in list(deps.keys()) :
						deps[curEdge] = Node(curEdge, curNode.depth+1)
			else :
				m = rSym.match (s)
				if m :
					symbols.append (m.group(1))
	# add the lost symbols found
	if curNode and curEdge and len(symbols) :
		curNode.AddEdge (curEdge, symbols, 0) # Todo: delay loaded
		symbols = []

	return deps

def main () :
	deps = {}
	dllsToRemove = []
	regexRemoves = []
	symbolsToRemove = []
	nMaxDepth = 10000 # almost unlimited
	components = []
	bDump = 0
	bDumpUnused = 0
	bDumpSymbolUse = 0
	bByUse = 0
	bReduce = 0
	bTred = 0
	bSaveDt = 0
	bSaveDsm = 0
	bSaveDB = 0
	bSaveXml = 0
	sOutFilename = None
	sPickle = None

	nSymbols = 0
	nCutLeafs = 0
	bRemoveNonLocal = 0
	bDumpReverse = 0

	if IsWin32() :
		# check if we are running from the right environment
		# HB: completely bogus, only works with vc2003?
		#try :
		#	print "FrameworkSDKDir=" + os.environ["FrameworkSDKDir"]
		#except :
		#	print "The script needs to be called from a 'VS command prompt'"
		#	sys.exit (1)
		if FindInPath ("dumpbin.exe") == "dumpbin.exe" :
			print("dumpbin.exe not found")
			sys.exit (1)
	for arg in sys.argv[1:] :
		if arg.find ("--remove") == 0 :
			if arg == "--remove-sys" : dllsToRemove.extend (dllsSysWin32)
			elif arg == "--remove-crt" : dllsToRemove.extend (dllsCrts)
			elif arg == "--remove-mfc" : dllsToRemove.extend (dllsMfc)
			elif arg == "--remove-gtk" : dllsToRemove.extend (dllsGtk)
			elif arg == "--remove-non-local" : bRemoveNonLocal = 1
			elif arg.find ("--remove-regex=") == 0 :
				regexRemoves.append (arg[len("--remove-regex="):])
			elif arg.find ("--remove-symbols=") == 0 :
				noSyms = CreateList(arg[len("--remove-symbols="):])
				symbolsToRemove.extend (noSyms)
			else :
				noDeps = CreateList(arg[len("--remove="):])
				dllsToRemove.extend(noDeps)
		elif arg.find ("--dont-follow") == 0 :
			global g_DontFollow
			noFollow = CreateList(arg[len("--dont-follow="):])
			g_DontFollow.extend (noFollow)
		elif arg.find ("--depth=") == 0 :
			nMaxDepth = int(arg[len("--depth="):])
			if nMaxDepth < 1 : print("Wrong depth"); sys.exit(1)
		elif arg.find ("--symbols=") == 0 :
			nSymbols = int(arg[len("--symbols="):])
			if nSymbols < 0 : nSymbols = 0
		elif arg.find ("--cut-leafs") == 0 :
			if arg.find ("--cut-leafs=") == 0 :
				nCutLeafs = int(arg[len("--cut-leafs="):])
			else :
				nCutLeafs = 10000 # infinite ;)
		elif arg == "--dump" :
			bDump = 1
		elif arg == "--dump-symbol-use" :
			bDumpSymbolUse = 1
			bDump = 1
		elif arg == "--dump-unused" :
			bDumpUnused = 1
			bDump = 1
		elif arg == "--dump-reverse" :
			bDumpReverse = 1
		elif arg == "--dt" :
			bSaveDt = 1
		elif arg == "--dsm" :
			bSaveDsm = 1
		elif arg == '--db' :
			bSaveDB = 1
		elif arg == "--xml" :
			bSaveXml = 1
		elif arg == "--reduce" :
			bReduce = 1
		elif arg == "--tred" :
			bTred = 1
		elif arg == "--by-use" :
			bByUse = 1
		elif arg.find ("--pickle=") == 0 :
			sGraph = arg[len("--pickle="):]
			sPickle = arg[len("--pickle="):] + ".pickle"
		elif arg.find ("--path") == 0 :
			global g_path
			if arg.find ("--path=") == 0 :
				g_path = arg[len("--path="):]
			else :
				g_path = os.environ['PATH']
		elif arg.find ("--theme=") == 0 :
			theme = arg[len("--theme="):]
			if theme == 'regex' :
				ParseRegexTheme('wdeps.tinting')
			else :
				print("Unknown theme!")
				sys.exit(1)
		elif arg.find ("--") == 0 :
			print("Unknown option or missing parameter:", arg)
			sys.exit(1)
		else :
			if len(components) == 0 :
				components = arg.split(",")
				components = UnGlob (components)
				# don't GetDeps here cause we need to first evaluate *all* parameters
				if len(components) == 0 :
					print("No DLL input found - wrong directory?")
					break
				sGraph = components[0]
			elif len(arg)>0 :
				sOutFilename = arg
		print("arg is:", arg)
	if len(sys.argv) < 2 :
		print(sys.argv[0], "[parameters] <component,s> [dot-output]")
		print("""
Given one or mores starting components like

	a.dll,another.dll

this tool anayzes the dependencies of the DLLs and generates a depency graph
in the dot format (see: www.graphviz.org). This tool requires DUMPBIN from
the MSVC toolchain. With the version included in VC7.1 it is capable to also
'see' 'delay load imports'. The tool follows the depencies to DLLs in the
current directory only.

If a second parameter is given the depency graph is written to that file.
Otherwise output is written to stdout.

There is also a number of extra parameters to fine-tune the graph.
First you can use --remove{-crt,-sys,-mfc} to remove the respective
set of dependencies. There is also a special form of remove
--remove=another.dll removes excatly that dll from the graph.

Another way to get more manageable graphs is to limit the recursion
depth while searching, like --depth=2 which cuts everything below
the second dependency level.

The most useful switch to anaylze parts of a highly coupled system is
--dont-follow=random.dll (accepting a list as above). This will show
random.dll in the resulting graph in gray. But it will not drag in
any dependencies below.

If you have a small graph it may be useful to not only show the
number of symbols used but their name. Giving --symbols=n all edges
with n or less symbols will have the (mostly demangled) name of the
symbols printed as text.

Finally there is the switch --dump for people prefering the textual
representation over some graph.

For more information read the source.
""")
		sys.exit(0)

	# output ...
	if not sOutFilename :
		f = sys.stdout
	else :
		# ... dot (or something)
		f = open(sOutFilename, "w")

	if bDump : # remember the command line
		f.write ("# " + " ".join (sys.argv) + "\n")

	try :
		import pickle
		deps = pickle.load (open(sPickle))
		print("Input from", sPickle)
	except :
		for s in components:
			if IsWin32() :
				GetDepsWin32 (s, deps, nMaxDepth)
			else :
				GetDepsPosix (s, deps, nMaxDepth)
		# with a bit dirty condition, try to populate from .dump
		if len(list(deps.keys())) == 1 and len(components) == 1:
			deps = ImportDump (components[0], deps)
			RecalculateDepth (deps)

	if len(dllsToRemove) :
		Remove (deps, dllsToRemove)
	if len(symbolsToRemove) > 0 :
		RemoveBySymbols (deps, symbolsToRemove)

	for rr in regexRemoves :
		RemoveRegEx (deps, rr)

	if bRemoveNonLocal :
		RemoveNonLocal (deps)

	while nCutLeafs > 0 :
		# not always iterating here, CutLeafs does too
		if bDump :
			nTotal = len(list(deps.keys()))
			leafs = CutLeafs (deps, 1)
			if len(leafs) < 1 :
				break
			f.write ("CutLeafs " + str(nTotal) + " => " + str(nTotal - len(leafs)) + "\n")
			leafs.sort()
			f.write ("\t" + ",".join (leafs) + "\n")
			nCutLeafs -= 1
		else :
			leafs = CutLeafs (deps, nCutLeafs)
			break

	if bDump :
		topmost = TopMost(deps)
		f.write ("TopMost(%d) : %s\n" % (len(topmost), ",".join(topmost)))

	# the *automatic* reduction is intentionally after cut-leafs
	if bReduce :
		if not sOutFilename :
			f2 = f
		else :
			f2 = open (sOutFilename + ".reduced.txt", "w")
		Reduce (deps, f2)
		if nCutLeafs > 1000 : # magic number two, still infinite ;)
			for d in list(deps.keys()) :
				# kind of arbitrary numbers
				n1 = len(list(deps[d].deps.keys())) # lthis ones dependencies
				n2 = len(list(deps.keys())) # total number of components left
				# write erros/warning like the compiler would do
				f2.write ("%s - %d error(s), %d warning(s)\n" % (deps[d].name, n1, n2))
	if bTred :
		tops = TopMost (deps)
		if len(tops) == 0 :
			print("Transitive reduction without topmost!")
			tops = list(deps.keys ())
		TransitiveReduction (deps, tops)

	if sPickle :
		import pickle
		pickle.dump(deps, open(sPickle,"w"))

	if bDumpSymbolUse :
		DumpSymbolsUse (deps, f)
		sys.exit (0)
	elif bDumpUnused :
		DumpUnusedSymbols (deps, f)
		sys.exit (0)
	elif bDump :
		DumpSymbols (deps, f)
		# no diagram at all
		sys.exit (0)
	if bSaveDt :
		SaveDt (deps, f)
	elif bSaveXml :
		SaveXml (deps, f)
	elif bSaveDsm :
		SaveDsm (deps, f)
	elif bSaveDB :
		SaveDB (deps, sOutFilename)
	elif bDumpReverse :
		DumpReverse (deps, f)
	else :
		SaveDot (deps, sGraph, bByUse, nSymbols, f)

def SaveXml (deps, f) :
	"simple XML dump"
	f.write('<nodes count="%d">\n' % (len(list(deps.keys()))))
	f.write('\t'*1 + '<command>wdeps.py ' + ' '.join (sys.argv[1:]) + '</command>\n')
	deps_keys = list(deps.keys())
	deps_keys.sort()
	for sn in deps_keys :
		node = deps[sn]
		f.write('\t'*1 + '<node name="%s" symbols="%d">\n' % (sn, len(node.exports)))
		edge_keys = list(node.deps.keys())
		if edge_keys :
			edge_keys.sort()
			f.write('\t'*2 + '<edges count="%d">\n' % (len(edge_keys)))
			for en in edge_keys :
				edge = node.deps[en]
				f.write('\t'*3 + '<edge name="%s" weight="%g">\n' % (en, edge.Weight()))
				f.write('\t'*4 + '<symbols>\n')
				symbols = edge.symbols
				symbols.sort()
				for sym in symbols :
					f.write('\t'*5 + '<symbol name="%s"/>\n' % (sym))
				f.write('\t'*4 + '</symbols>\n')
				f.write('\t'*3 + '</edge>\n')
			f.write('\t'*2 + '</edges>')
		f.write('\t'*1 + '</node>\n')
	f.write('</nodes>\n')

def SaveDt (deps, f) :
	""" see: http://dtangler.org """
	deps_keys = list(deps.keys())
	deps_keys.sort()
	for sn in deps_keys :
		node = deps[sn]
		edge_keys = list(node.deps.keys())
		if not edge_keys :
			continue
		edge_keys.sort()
		f.write (sn + " : " + " ".join (edge_keys) + "\n")

def SaveDot (deps, sGraph, bByUse, nSymbols, f) :
	# build the graph
	f.write ('digraph "' + sGraph + '" {\n')
	f.write ('graph [fontsize=8.0 label="wdeps.py ' + " ".join (sys.argv[1:])
			+ '\\n' + time.ctime() + '"]\n')
	f.write ('ratio=0.7\nnode [fontsize=12.0 ]\nedge [fontsize=8.0]\n')
	if bByUse :
		# kind of inverted diagram not showing edependencies but 'users'
		users = {}
		for sn in list(deps.keys()) :
			users[sn] = []
		for sn in list(deps.keys()) :
			node = deps[sn]
			for se in list(node.deps.keys()) :
				edge = node.deps[se]
				users[edge.name].append(node.name)
		for sn in list(users.keys()) :
			for s in users[sn] :
				f.write ('"%s" -> "%s"\n' % (sn, s))
	else :
		# first pass mark don't follows
		dontFollowsDone = {}
		urlsDone = {}
		deps_keys = list(deps.keys())
		deps_keys.sort()
		for sn in deps_keys :
			node = deps[sn]
			edge_keys = list(node.deps.keys())
			edge_keys.sort()
			for se in edge_keys :
				cut_len = len(se)
				if IsWin32() :
					cut_len = se.find(".dll")
				else :
					cut_len = se.find(".so")
				ses = se[:cut_len]
				if se in g_DontFollow :
					if se not in dontFollowsDone :
						# mark as such
						f.write ('"%s" [style=filled,fillcolor="lightgray"]\n' % (se,))
						dontFollowsDone[se] = 1
				else :
					if se not in urlsDone :
						f.write ('"%s" [style=filled,fillcolor="%s",URL="#%s"]\n' % (se, LookupColor(deps[se]), se))
						urlsDone[se] = 1


		for sn in deps_keys :
			# write weighted edges, could also classify the nodes ...
			node = deps[sn]
			edge_keys = list(node.deps.keys())
			edge_keys.sort()
			for se in edge_keys :
				edge = node.deps[se]
				sPrefix = ""
				if edge.reduce : # remove from graph by commenting it out
					sPrefix = "// " + ", ".join(edge.symbols) + "\n// "
				if edge.delayLoad :
					# putting 'weight' and 'constraint' seems to be too much for dot
					sStyle = "weight=%f,style=dotted" % (math.log10(math.sqrt(edge.Weight()+.1)),)
					# even using constraint at all seems to often crash dot
					#sStyle = "style=dotted,constraint=false"
				else :
					sStyle = "weight=%f" % (math.log10(edge.Weight()+.1),)
				if edge.Weight() <= nSymbols :
					#f.write ('"%s" -> "%s" [weight=%f,label=%s]\n' % (node.name, edge.name, math.log(1)-0.5, edge.symbols[0]))
					f.write ('%s"%s" -> "%s" [fontsize=6,label="%s",%s]\n'
							% (sPrefix, node.name, edge.name, "\\n".join(edge.symbols), sStyle))
				else :
					#f.write ('"%s" -> "%s" [weight=%f]\n' % (node.name, edge.name, math.log(edge.Weight())-0.5))
					f.write ('%s"%s" -> "%s" [label="(%d)",%s]\n'
							% (sPrefix, node.name, edge.name, edge.Weight(), sStyle))
	f.write("}\n")

if __name__ == '__main__': main()



#!/usr/bin/env python

#   Interactive Python-GTK Console
#   Copyright (C), 1998 James Henstridge <james@daa.com.au>
#
#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

# This module implements an interactive python session in a GTK window.  To
# start the session, use the gtk_console command.  Its specification is:
#   gtk_console(namespace, title, copyright)
# where namespace is a dictionary representing the namespace of the session,
#       title     is the title on the window and
#       copyright is any additional copyright info to print.
#
# As well as the starting attributes in namespace, the session will also
# have access to the list __history__, which is the command history.

import warnings
import sys, traceback

import gi

gi.require_version('Pango', '1.0')
gi.require_version('Gtk', '3.0')

with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=RuntimeWarning)
    from gi.repository import GObject, Pango, Gtk, Gdk

import gettext

_ = gettext.gettext

stdout = sys.stdout

if not hasattr(sys, "ps1"):
    sys.ps1 = ">>> "
if not hasattr(sys, "ps2"):
    sys.ps2 = "... "

# some functions to help recognise breaks between commands
def remQuotStr(s):
    """Returns s with any quoted strings removed (leaving quote marks)"""
    r = ""
    inq = 0
    qt = ""
    prev = "_"
    while len(s):
        s0, s = s[0], s[1:]
        if inq and (s0 != qt or prev == "\\"):
            prev = s0
            continue
        prev = s0
        if s0 in "'\"":
            if inq:
                inq = 0
            else:
                inq = 1
                qt = s0
        r = r + s0
    return r


def bracketsBalanced(s):
    """Returns true iff the brackets in s are balanced"""
    s = [x for x in s if x in "()[]{}"]
    stack = []
    brackets = {"(": ")", "[": "]", "{": "}"}
    while len(s) != 0:
        if s[0] in ")]}":
            if len(stack) != 0 and brackets[stack[-1]] == s[0]:
                del stack[-1]
            else:
                return 0
        else:
            stack.append(s[0])
        s = s[1:]
    return len(stack) == 0


class gtkoutfile:
    """A fake output file object.  It sends output to a TK test widget,
    and if asked for a file number, returns one set on instance creation"""

    def __init__(self, w, fn, tag):
        self.__fn = fn
        self.__w = w
        self.__tag = tag

    def close(self):
        pass

    flush = close

    def fileno(self):
        return self.__fn

    def isatty(self):
        return 0

    def read(self, a):
        return ""

    def readline(self):
        return ""

    def readlines(self):
        return []

    def write(self, s):
        # stdout.write(str(self.__w.get_point()) + '\n')
        self.__w.text_insert(self.__tag, s)

    def writelines(self, l):
        self.__w.text_insert(self.__tag, l)

    def seek(self, a):
        raise IOError(29, "Illegal seek")

    def tell(self):
        raise IOError(29, "Illegal seek")

    truncate = tell


class Console(Gtk.VBox):
    def __init__(self, namespace={}, copyright="", quit_cb=None):
        GObject.GObject.__init__(self, spacing=2)
        self.copyright = copyright

        self.quit_cb = quit_cb

        self.inp = Gtk.HBox()
        self.pack_start(self.inp, True, True, 0)
        self.inp.show()

        self.scroll = Gtk.ScrolledWindow()
        self.inp.pack_start(self.scroll, True, True, 0)
        self.scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.ALWAYS)
        self.scroll.show()

        self.text = Gtk.TextView()
        self.text.modify_font(Pango.FontDescription("monospace 8"))
        self.text.set_editable(False)
        self.text.set_wrap_mode(Gtk.WrapMode.WORD)
        self.text.set_size_request(500, 400)
        self.scroll.add(self.text)
        self.text.show()

        buffer = self.text.get_buffer()
        self.normal = buffer.create_tag("normal")
        self.title = buffer.create_tag("title")
        self.title.set_property("weight", 600)
        self.error = buffer.create_tag("error")
        self.error.set_property("foreground", "red")
        self.command = buffer.create_tag("command")
        self.command.set_property("weight", 600)
        self.command.set_property("foreground", "blue")

        self.inputbox = Gtk.HBox(spacing=2)
        self.pack_end(self.inputbox, False, True, 0)
        self.inputbox.show()

        self.prompt = Gtk.Label(label=sys.ps1)
        self.prompt.set_padding(2, 0)
        self.prompt.set_size_request(26, -1)
        self.inputbox.pack_start(self.prompt, False, False, 0)
        self.prompt.show()

        self.line = Gtk.Entry()
        self.line.set_size_request(400, -1)
        self.line.connect("key_press_event", self.key_function)
        self.inputbox.pack_start(self.line, True, True, 2)
        self.line.show()

        # now let the text box be resized
        self.text.set_size_request(0, 0)
        self.line.set_size_request(0, -1)

        self.namespace = namespace

        self.cmd = ""
        self.cmd2 = ""

        # set up hooks for standard output.
        self.stdout = gtkoutfile(self, sys.stdout.fileno(), self.normal)
        try:
            # this will mostly fail on win32 ...
            self.stderr = gtkoutfile(self, sys.stderr.fileno(), self.error)
        except:
            # ... but gtkoutfile is not using the fileno anyway
            self.stderr = gtkoutfile(self, -1, self.error)
        # set up command history
        self.history = [""]
        self.histpos = 0
        self.namespace["__history__"] = self.history

    def scroll_to_end(self):
        iter = self.text.get_buffer().get_end_iter()
        self.text.scroll_to_iter(iter, 0, False, 0, 0)
        return False  # don't requeue this handler

    def text_insert(self, tag, s):
        buffer = self.text.get_buffer()
        iter = buffer.get_end_iter()
        buffer.insert_with_tags(iter, s, tag)
        GObject.idle_add(self.scroll_to_end)

    def init(self):
        self.text.realize()
        # ??? self.text.style = self.text.get_style()
        # self.text.fg = self.text.style.fg[Gtk.StateType.NORMAL]
        # self.text.bg = self.text.style.white

        self.text_insert(
            self.title,
            "Python %s\n%s\n\n" % (sys.version, sys.copyright)
            + "Interactive Python-GTK Console - "
            + "Copyright (C) 1998 James Henstridge\n\n"
            + self.copyright
            + "\n",
        )
        self.line.grab_focus()

    def quit(self, *args):
        self.destroy()
        if self.quit_cb:
            self.quit_cb()

    def key_function(self, entry, event):
        if event.keyval == Gdk.KEY_Return:
            self.line.emit_stop_by_name("key_press_event")
            self.eval()
        if event.keyval == Gdk.KEY_Tab:
            self.line.emit_stop_by_name("key_press_event")
            self.line.append_text("\t")
            GObject.idle_add(self.focus_text)
        elif event.keyval in (Gdk.KEY_KP_Up, Gdk.KEY_Up):
            self.line.emit_stop_by_name("key_press_event")
            self.historyUp()
            GObject.idle_add(self.focus_text)
        elif event.keyval in (Gdk.KEY_KP_Down, Gdk.KEY_Down):
            self.line.emit_stop_by_name("key_press_event")
            self.historyDown()
            GObject.idle_add(self.focus_text)
        elif (
            event.keyval in (Gdk.KEY_D, Gdk.KEY_d)
            and event.get_state() & Gdk.ModifierType.CONTROL_MASK
        ):
            self.line.emit_stop_by_name("key_press_event")
            self.ctrld()

    def focus_text(self):
        self.line.grab_focus()
        return False  # don't requeue this handler

    def ctrld(self):
        self.quit()
        pass

    def historyUp(self):
        if self.histpos > 0:
            l = self.line.get_text()
            if len(l) > 0 and l[0] == "\n":
                l = l[1:]
            if len(l) > 0 and l[-1] == "\n":
                l = l[:-1]
            self.history[self.histpos] = l
            self.histpos = self.histpos - 1
            self.line.set_text(self.history[self.histpos])

    def historyDown(self):
        if self.histpos < len(self.history) - 1:
            l = self.line.get_text()
            if len(l) > 0 and l[0] == "\n":
                l = l[1:]
            if len(l) > 0 and l[-1] == "\n":
                l = l[:-1]
            self.history[self.histpos] = l
            self.histpos = self.histpos + 1
            self.line.set_text(self.history[self.histpos])

    def eval(self):
        l = self.line.get_text() + "\n"
        if len(l) > 1 and l[0] == "\n":
            l = l[1:]
        self.histpos = len(self.history) - 1
        if len(l) > 0 and l[-1] == "\n":
            self.history[self.histpos] = l[:-1]
        else:
            self.history[self.histpos] = l
        self.line.set_text("")
        self.text_insert(self.command, self.prompt.get_label() + l)
        if l == "\n":
            self.run(self.cmd)
            self.cmd = ""
            self.cmd2 = ""
            return
        self.histpos = self.histpos + 1
        self.history.append("")
        self.cmd = self.cmd + l
        self.cmd2 = self.cmd2 + remQuotStr(l)
        l = l.rstrip()
        if (
            not bracketsBalanced(self.cmd2)
            or l[-1] == ":"
            or l[-1] == "\\"
            or l[0] in " \11"
        ):
            self.prompt.set_text(sys.ps2)
            self.prompt.queue_draw()
            return
        self.run(self.cmd)
        self.cmd = ""
        self.cmd2 = ""

    def run(self, cmd):
        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr
        try:
            try:
                r = eval(cmd, self.namespace, self.namespace)
                if r is not None:
                    print(repr(r))
            except SyntaxError:
                exec(cmd, self.namespace)
        except:
            if hasattr(sys, "last_type") and sys.last_type == SystemExit:
                self.quit()
            else:
                traceback.print_exc()
        self.prompt.set_text(sys.ps1)
        self.prompt.queue_draw()
        # adj = self.text.get_vadjustment()
        # adj.set_value(adj.upper - adj.page_size)
        sys.stdout, self.stdout = self.stdout, sys.stdout
        sys.stderr, self.stderr = self.stderr, sys.stderr


def gtk_console(ns, title="Python", copyright="", menu=None):
    win = Gtk.Window()
    win.set_default_size(475, 300)
    # win.connect("destroy", mainquit)
    win.set_title(title)

    def quit(win=win):
        win.destroy()

    cons = Console(namespace=ns, copyright=copyright, quit_cb=quit)
    if menu:
        box = Gtk.VBox()
        win.add(box)
        box.show()
        box.pack_start(menu, False, True, 0)
        menu.show()
        box.pack_start(cons, True, True, 0)
    else:
        win.add(cons)
    cons.show()
    win.show()
    cons.init()


# set up as a dia plugin
try:
    import dia

    def open_console(data, flags):
        gtk_console(
            {
                "__builtins__": __builtins__,
                "__name__": "__main__",
                "__doc__": None,
                "dia": dia,
            },
            _("Python Dia Console"),
        )

    dia.register_action(
        "DialogsPythonconsole",
        _("_Python Console"),
        "/DisplayMenu/Dialogs/DialogsExtensionStart",
        open_console,
    )

except:
    print("Failed to import Dia ...")
    gtk_console({"__builtins__": __builtins__, "__name__": "__main__", "__doc__": None})
    Gtk.main()



#    Select Objects By Common Properties
#    Copyright (c) 2003, 2013 Hans Breuer <hans@breuer.org>
#
#        different methods (menu entries) to select all objects with given
#    property, that is :
#	- a find dialog to select by name
#	- menu entries for different color types. The common color is read
#	  from already selected objects
#	- a size range given by the current selection

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import warnings
import dia

import gettext
_ = gettext.gettext

class CFindDialog :
	def __init__(self, data) :
		import gi

		gi.require_version('Gtk', '3.0')

		with warnings.catch_warnings():
			warnings.filterwarnings("ignore", category=RuntimeWarning)
			from gi.repository import Gtk

		win = Gtk.Window()
		win.connect("delete_event", self.on_delete)
		win.set_title(_("Select by name"))

		self.diagram = data
		self.win = win

		box1 = Gtk.VBox()
		win.add(box1)
		box1.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, True, True, 0)
		box2.show()

		self.entry = Gtk.Entry()
		self.entry.set_text(_("Enter name"))
		box2.pack_start(self.entry, True, True, 0)
		self.entry.show()

		separator = Gtk.HSeparator()
		box1.pack_start(separator, 0, True, 0)
		separator.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, 0, True, 0)
		box2.show()

		button = Gtk.Button(_("Find"))
		button.connect("clicked", self.on_find)
		box2.pack_start(button, True, True, 0)
		button.set_can_default(True)
		button.grab_default()
		button.show()
		win.show()

	def on_find(self, *args) :
		s = self.entry.get_text()
		select_by (self.diagram, "name", s)
		self.diagram.update_extents ()
		self.diagram.flush()

	def on_delete (self, *args) :
		self.win.destroy ()

def do_dialog(data):
	# let the dialog do it's work
	dlg = CFindDialog(data)

def select_by (diagram, name, value) :
	objs = diagram.active_layer.objects
	for o in objs :
		if name in o.properties :
			print("<==", o.properties[name].value)
			if value == None or o.properties[name].value == value :
				diagram.select (o)

def select_by_name_cb (data, flags):
	do_dialog (data)

def select_by_selected (data, name) :
	grp = data.get_sorted_selected()
	bFoundAny = 0
	for o in grp :
		if name in o.properties :
			select_by (data, name, o.properties[name].value)
			bFoundAny = 1
	if not bFoundAny :
		dia.message(0, "No selected object has the property '%s'." % name)
	data.update_extents ()
	data.flush()

def select_by_fill_color_cb (data, flags) :
	select_by_selected (data, "fill_colour")
def select_by_line_color_cb (data, flags) :
	select_by_selected (data, "line_colour")
def select_by_text_color_cb (data, flags) :
	select_by_selected (data, "text_colour")

def select_by_size_cb (data, flags) :
	""" From the diagrams selection derive the minimum and maximum object size.
	    Select every other object within this range.
	"""
	grp = data.get_sorted_selected()
	smin = 100000
	smax = 0
	for o in grp :
		w = o.bounding_box.right - o.bounding_box.left
		h = o.bounding_box.bottom - o.bounding_box.top
		s = w * h
		if s > smax :
			smax = s
		if s < smin :
			smin = s
	objs = data.active_layer.objects
	for o in objs :
		w = o.bounding_box.right - o.bounding_box.left
		h = o.bounding_box.bottom - o.bounding_box.top
		s = w * h
		if s >= smin and s <= smax :
			data.select(o)
	data.flush()


dia.register_action ("SelectByName", _("_Name"),
                       "/DisplayMenu/Select/SelectBy/SelectByExtensionStart",
                       select_by_name_cb)
dia.register_action ("SelectByFillcolor", _("_Fill Color"),
                       "/DisplayMenu/Select/SelectBy/SelectByExtensionStart",
                       select_by_fill_color_cb)
dia.register_action ("SelectByLinecolor", _("_Line Color"),
                       "/DisplayMenu/Select/SelectBy/SelectByExtensionStart",
                       select_by_line_color_cb)
dia.register_action ("SelectByTextcolor", _("_Text Color"),
                       "/DisplayMenu/Select/SelectBy/SelectByExtensionStart",
                       select_by_text_color_cb)
dia.register_action ("SelectBySize", _("_Size"),
                       "/DisplayMenu/Select/SelectBy/SelectByExtensionStart",
                       select_by_size_cb)



import sys, os
import dia

# Please don't include pygtk or gtk here. There
# are quite some PyDia plug-ins which don't require
# i.e. dont have any (Py)Gtk code

loaded = []

def load(plugindir):
	global loaded
	if os.path.isdir(plugindir):
		sys.path.insert(0, plugindir)
		for file in os.listdir(plugindir):
			if file[-3:] == '.py' and file[:-3] not in loaded:
				try:
					__import__(file[:-3])
					loaded.append(file[:-3])
				except Exception as e:
					sys.stderr.write('could not import %s [%s]\n' % (file, e))

# import any python plugins from the user ...
if 'HOME' not in os.environ:
	os.environ['HOME'] = os.pathsep + 'tmp'
# import all plugins found in user plugin dir
load(os.path.join(os.environ['HOME'], '.dia', 'python'))

# find system python plugin dir
curdir = os.path.dirname(__file__)
plugindir = os.path.join(curdir, 'python')
#If we're running from source there is no plugin dir
if not os.path.exists(plugindir):
	plugindir = curdir
load(plugindir)



# Copyright (c) 2007  Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
import dia
import math

import gettext
_ = gettext.gettext

def DeepCalc (dict, key, seen = None) :
	# calculate all (deep) dependencies
	# give a directory with component name as key and dependency components as list in dict[key][2]
	if not seen :
		seen = {}
	for k in dict[key][2] :
		if seen and k in seen:
			continue
		seen[k] = 1
		DeepCalc (dict, k, seen)
	return len(list(seen.keys()))

##
# \brief Callback function to be invoked by Dia's menu
#
# Implements a simple layout algorithm based on object connections
#
def arrange_connected (data, flags) :
	objs = data.get_sorted_selected()
	if len(objs) == 0 :
		objs = data.active_layer.objects
	# all objects having at least one connection
	bs = {}
	print("objs", len(objs))
	edges = {}
	for o in objs :
		for c in o.connections: # ConnectionPoint
			for n in c.connected: # connection object
				if not (n.handles[0].connected_to and n.handles[1].connected_to):
					continue
				a = n.handles[0].connected_to.object
				b = n.handles[1].connected_to.object
				ak = a.properties["name"].value
				bk = b.properties["name"].value
				# create an edge key
				ek = ak + "->" + bk
				if ek in edges:
					continue # already seen
				print(ek)
				edges[ek] = 1
				if ak in bs:
					use = bs[ak]
					use[2].append (bk)
					bs[ak] = use
				else :
					bs[ak] = [a, 0, [bk]]
				if bk in bs:
					use = bs[bk]
					use[1] += 1
					bs[bk] = use
				else :
					bs[bk] = [b, 1, []]
	# sort by number of connections
	bst = []
	dx = 0
	dy = 0
	for key in list(bs.keys()) :
		o = bs[key][0]
		if not o.properties["elem_width"] :
			print(o)
			continue
		if o.properties["elem_width"].value > dx : dx = o.properties["elem_width"].value
		if o.properties["elem_height"].value > dy : dy = o.properties["elem_height"].value
		n = bs[key][1] # (use count, dependencies)
		bst.append ((o, n, DeepCalc (bs, key)))
	bst.sort(key=lambda a: a[1])
	if len(bs) < 2 :
		return
	# average weight gives the number of rows
	aw = 0.0
	for t in bst :
		aw += (t[1] + t[2])
	aw = aw / len(bst)
	rows = int (len(bst) / math.sqrt(aw))
	if rows < 2 :
		rows = 2
	offsets = []
	for i in range(0, rows) : offsets.append(0)
	dx *= 1.4
	dy *= 1.8
	for t in bst :
		y = t[1] # t[1] = 0 :  start, noone uses it
		# correction to move further down
		c = rows - (rows * t[2] / aw)
		if t[1] : # the less users the far below
			if t[2] == 0 :
				y = rows - 1
			elif t[1] >= rows :
				# compensate for some of the weight, FIXME: better guess needed
				y = int(rows - 2 * aw / (t[2] + t[1]))
		print(t[0].properties["name"].value, t[1], t[2], y, c)
		# move the object to it's new place
		x = offsets[y]
		t[0].move (x * dx, y * dy)
		offsets[y] += 1
	data.update_extents ()

##
# \file arrange.py \brief Arrange Objects Plugin
#
# this module is loaded by some other plug-ins but can also work on it's own
# if it is loaded first as Dia plug-in and later as Python module everything works
# fine due to Pythoninitializing the module only once
#
# \ingroup PyDia
dia.register_callback (_("Arrange _Objects"),
                       "<Display>/Objects/Arrange",
                       arrange_connected)



#    PyDia Self Documentation Series - Part III : All Objects
#    Copyright (c) 2007, Hans Breuer <hans@breuer.org>
#
#        generates a new diagram which contains all the currently
#    registered objects sorted to layers by their containing package
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia

import gettext
_ = gettext.gettext

def set_object_string (o) :
	keys = list(o.properties.keys())
	for s in keys :
		p = o.properties[s]
		if p.type in ["string", "text"] :
			if s in ["name", "text"] :
				o.properties[s] = o.type.name
			else :
				o.properties[s] = s

def aobjects_cb(data, flags) :

	# copied from otypes.py
	if data:
		diagram = None # we may be running w/o GUI
	else :
		diagram = dia.new("All Objects.dia")
		data = diagram
	layer = data.active_layer

	otypes = dia.registered_types()
	keys = list(otypes.keys())
	keys.sort()

	packages = {}
	for s in keys :
		kt = " - ".split(s)
		if len(kt) == 2 :
			if len(kt[0]) == 0 :
				sp = "<unnamed>"
			else :
				sp = kt[0]
			st = kt[1]
		else :
			sp = "<broken>"
			st = kt[0]
		if sp in packages:
			packages[sp].append(s)
		else :
			packages[sp] = [s]

	for sp in list(packages.keys()):
		# add a layer per package
		layer = data.add_layer (sp)

		cx = 0.0
		cy = 0.0
		n = 0 # counting objects
		my = 0.0
		pkg = packages[sp]
		for st in pkg :
			if st == "Group" :
				continue # can't create empty group
			#print st
			o, h1, h2 = dia.get_object_type(st).create (cx, cy)
			# to make the resulting diagram more interesting we set every sting property with it's name
			set_object_string (o)
			w = o.bounding_box.right - o.bounding_box.left
			h = o.bounding_box.bottom - o.bounding_box.top
			o.move (cx, cy)
			cx += w * 1.5
			if h > my : my = h
			n += 1
			if n % 10 == 0 :
				cx = 0.0
				cy += my * 1.5
				my = 0
			layer.add_object (o)
		layer.update_extents()
	data.update_extents()
	if diagram :
		diagram.display()
		diagram.flush()
	return data

dia.register_action ("HelpAObjects", _("All _Objects"),
                     "/ToolboxMenu/Help/HelpExtensionStart",
                     aobjects_cb)



#  PyDia SVG Import
#  Copyright (c) 2003, 2004 Hans Breuer <hans@breuer.org>
#
#  Pure Python Dia Import Filter - to show how it is done.
#  It also tries to be more featureful and robust then the
#  SVG importer written in C, but as long as PyDia has issues
#  this will _not_ be the case. Known issues (at least) :
#  - xlink stuff (should probably have some StdProp equivalent)
#  - lack of full transformation dealing
#  - real percentage scaling, is it worth it ?
#  - see FIXME in this file

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import string, math, os, re

import gettext
_ = gettext.gettext

# Dias unit is cm, the default scale should be determined from svg:width and viewBox
dfPcm = 35.43307
dfUserScale = 1.0
dfFontSize = 0.7
dfViewLength = 32.0 # wrong approach for "% unit"
dictUnitScales = {
	"em" : 1.0, "ex" : 2.0, #FIXME these should be _relative_ to current font
	"px" : 1.0 / dfPcm, "pt" : 1.25 / dfPcm, "pc" : 15.0 / dfPcm,
	"cm" : 35.43307 / dfPcm, "mm" : 3.543307 / dfPcm, "in" : 90.0 / dfPcm}

# only compile once
rColor = re.compile(r"rgb\s*\(\s*(\d+)[, ]+(\d+)[, +](\d+)\s*\)")
# not really parsing numbers (Scaled will deal with more)
rTranslate = re.compile(r"translate\s*\(\s*([^,]+),([^)]+)\s*\)")
#FIXME: parse more - e.g. AQT - of the strange path data
rPathWhat = re.compile("[MmLlCcSsZz]")   # what
rPathData = re.compile("[^MmLlCcSsZz]+") # data
rPathValue = re.compile("[\s,]+") # values

def Scaled(s) :
	# em, ex, px, pt, pc, cm, mm, in, and percentages
	if s[-1] in string.digits :
		# use global scale
		return float(s) * dfUserScale
	else :
		unit = s[-2:]
		try :
			if unit[0] == "e" :
				#print "Scaling", unit, dfFontSize
				return float(s[:-2]) * dfFontSize * dictUnitScales[unit]
			else :
				return float(s[:-2]) * dictUnitScales[unit]
		except :
			if s[-1] == "%" :
				return float(s[:-1]) * dfViewLength / 100.0
			# warn about invalid unit ??
			raise NotImplementedError("Unknown unit %s %s" % (s[:-2], s[-2:]))
			return float(s) * dfUserScale
def Color(s) :
	# deliver a StdProp compatible Color (or the original string)
	m = rColor.match(s)
	if m :
		return (int(m.group(1)) / 255.0, int(m.group(2)) / 255.0, int(m.group(3)) / 255.0, 1.0)
	# any more ugly color definitions not compatible with pango_color_parse() ?
	return s.strip()
def _eval (s, _locals) :
	# eval() can be used to execute aribitray code, see e.g. http://bugzilla.gnome.org/show_bug.cgi?id=317637
	# here using *any* builtins is an abuse
	try :
		return eval (s, {'__builtins__' : None }, _locals)
	except NameError :
		try :
			import dia
			dia.message(2, "***Possible exploit attempt***:\n" + s)
		except ImportError :
			print("***Possible exploit***:", s)
	return None
class Object :
	def __init__(self) :
		self.props = {"x" : 0, "y" : 0, "stroke" : "none"}
		self.translation = None
		# "line_width", "line_colour", "line_style"
	def style(self, s) :
		sp1 = ";".split(s)
		for s1 in sp1 :
			sp2 = s1.strip().split(":")
			if len(sp2) == 2 :
				try :
					_eval("self." + sp2[0].replace("-", "_") + "(\"" + sp2[1].strip() + "\")", locals())
				except AttributeError :
					self.props[sp2[0]] = sp2[1].strip()
	def x(self, s) :
		self.props["x"] = Scaled(s)
	def y(self, s) :
		self.props["y"] = Scaled(s)
	def width(self, s) :
		self.props["width"] = Scaled(s)
	def height(self, s) :
		self.props["height"] = Scaled(s)
	def stroke(self,s) :
		self.props["stroke"] = s.encode("UTF-8")
	def stroke_width(self,s) :
		self.props["stroke-width"] = Scaled(s)
	def fill(self,s) :
		self.props["fill"] = s
	def fill_rule(self,s) :
		self.props["fill-rule"] = s
	def stroke_dasharray(self,s) :
		# just an approximation
		sp = s.split(",")
		n = len(sp)
		if n > 0 :
			# sp[0] == "none" : # ? stupid generator ?
			try :
				dlen = Scaled(sp[0])
			except :
				n = 0
		if n == 0 : # should not really happen
			self.props["line-style"] = (0, 1.0) # DIA_LINE_STYLE_SOLID,
		elif n == 2 :
			if dlen > 0.1 : # FIXME:
				self.props["line-style"] = (1, dlen) # DIA_LINE_STYLE_DASHED,
			else :
				self.props["line-style"] = (4, dlen) # DIA_LINE_STYLE_DOTTED
		elif n == 4 :
			self.props["line-style"] = (2, dlen) # DIA_LINE_STYLE_DASH_DOT,
		elif n == 6 :
			self.props["line-style"] = (3, dlen) # DIA_LINE_STYLE_DASH_DOT_DOT,
	def id(self, s) :
		# just to handle/ignore it
		self.props["meta"] = { "id" : s }
	def transform(self, s) :
		m = rTranslate.match(s)
		if m :
			#print "matched", m.group(1), m.group(2), "->", Scaled(m.group(1)), Scaled(m.group(2))
			self.translation = (Scaled(m.group(1)), Scaled(m.group(2)))
	def __repr__(self) :
		return self.dt + " : " + str(self.props)
	def Dump(self, indent) :
		print(" " * indent, self)
	def Set(self, d) :
		pass
	def ApplyProps(self, o) :
		pass
	def CopyProps(self, dest) :
		# to be used to inherit group props to childs _before_ they get their own
		# doesn't use the member functions to avoid scaling once more
		for p in list(self.props.keys()) :
			dest.props[p] = self.props[p]
	def Create(self) :
		ot = dia.get_object_type (self.dt)
		o, h1, h2 = ot.create(self.props["x"], self.props["y"])
		# apply common props
		if "stroke-width" in self.props and "line_width" in o.properties :
			o.properties["line_width"] = self.props["stroke-width"]
		if "stroke" in self.props and "line_colour" in o.properties :
			if self.props["stroke"] != "none" :
				try :
					o.properties["line_colour"] = Color(self.props["stroke"])
				except :
					# rgb(192,27,38) handled by Color() but ...
					# o.properties["line_colour"] = self.props["stroke"]
					pass
			else :
				# Dia can't really display stroke none, some workaround :
				if "fill" in self.props and self.props["fill"] != "none" :
					#does it really matter ?
					try :
						o.properties["line_colour"] = Color(self.props["fill"])
					except :
						pass
				o.properties["line_width"] = 0.0
		if "fill" in self.props and "fill_colour" in o.properties :
			if self.props["fill"] == "none" :
				o.properties["show_background"] = 0
			else :
				color_key = "fill_colour"
				try :
					o.properties["show_background"] = 1
				except KeyError :
					# not sure if this is always true
					color_key = "text_colour"
				try :
					o.properties[color_key] =Color(self.props["fill"])
				except :
					# rgb(192,27,38) handled by Color() but ...
					# o.properties["fill_colour"] =self.props["fill"]
					pass
		if "line-style" in self.props and "line_style" in o.properties :
			o.properties["line_style"] = self.props["line-style"]
		if "meta" in self.props and "meta" in o.properties :
			o.properties["meta"] = self.props["meta"]
		self.ApplyProps(o)
		return o

class Svg(Object) :
	# not a placeable object but similar while parsing
	def __init__(self) :
		Object.__init__(self)
		self.dt = "svg"
		self.bbox_w = None
		self.bbox_h = None
	def width(self,s) :
		global dfUserScale
		d = dfUserScale
		dfUserScale = 0.05
		self.bbox_w = Scaled(s)
		self.props["width"] = self.bbox_w
		dfUserScale = d
	def height(self,s) :
		global dfUserScale
		d = dfUserScale
		# with stupid info Dia still has a problem cause zooming is limited to 5.0%
		dfUserScale = 0.05
		self.bbox_h = Scaled(s)
		self.props["height"] = self.bbox_h
		dfUserScale = d
	def viewBox(self,s) :
		global dfUserScale
		global dfViewLength
		self.props["viewBox"] = s
		sp = s.split(" ")
		w = float(sp[2]) - float(sp[0])
		h = float(sp[3]) - float(sp[1])
		# FIXME: the following relies on the call order of width,height,viewBox
		# which is _not_ the order it is in the file
		if self.bbox_w and self.bbox_h :
			dfUserScale = math.sqrt((self.bbox_w / w)*(self.bbox_h / h))
		elif self.bbox_w :
			dfUserScale = self.bbox_w / w
		elif self.bbox_h :
			dfUserScale = self.bbox_h / h
		# FIXME: ugly, simple aproach to "%" unit
		dfViewLength = math.sqrt(w*h)
	def xmlns(self,s) :
		self.props["xmlns"] = s
	def version(self,s) :
		self.props["version"] = s
	def __repr__(self) :
		global dfUserScale
		return Object.__repr__(self) + "\nUserScale : " + str(dfUserScale)
	def Create(self) :
		return None
class Style(Object) :
	# the beginning of a css implementation, ...
	def __init__(self) :
		global cssStyle
		Object.__init__(self)
		self.cdata = ""
		self.styles = None
		cssStyle = self
	def type(self, s) :
		self.props["type"] = s
	def Set(self, d) :
		# consuming all the ugly CDATA
		self.cdata += d
	def Lookup(self, st) :
		if self.styles == None :
			self.styles = {}
			# just to check if we are interpreting correctly (better use regex ?)
			p1 = 0 # position of dot
			p2 = 0 # ... of opening brace
			p3 = 0 # ... closing
			s = self.cdata
			n = len(s) - 1
			while True:
				p1 = s.find(".", p3, n)
				p2 = s.find("{", p1+1, n)
				p3 = s.find("}", p2+1, n)
				if p1 < 0 or p2 < 0 or p3 < 0 :
					break
				print(s[p1+1:p2-1], s[p2+1:p3])
				self.styles[s[p1+1:p2-1]] = s[p2+1:p3]
		if st in self.styles :
			return self.styles[st]
		return ""
	def __repr__(self) :
		self.Lookup("init") # fill the dictionary
		return "Styles:" + str(self.styles)
	def Create(self) :
		return None

cssStyle = Style() # a singleton

class Group(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Group"
		self.childs = []
	def Add(self, o) :
		self.childs.append(o)
	def Create(self) :
		lst = []
		for o in self.childs :
			od = o.Create()
			if od :
				#print od
				#DON'T : layer.add_object(od)
				lst.append(od)
		# create group including list objects
		if len(lst) > 0 :
			grp = dia.group_create(lst)
			if self.translation :
				# want to move by top left corner ...
				hNW = grp.handles[0] # HANDLE_RESIZE_NW
				# ... but pos is the point moved
				pos = grp.properties["obj_pos"].value
				#FIXME:  looking at scascale.py this isn't completely correct
				x1 = hNW.pos.x + self.translation[0]
				y1 = hNW.pos.y + self.translation[1]
				grp.move(x1, y1)
			return grp
		else :
			return None
	def Dump(self, indent) :
		print(" " * indent, self)
		for o in self.childs :
			o.Dump(indent + 1)

# One of my test files is quite ugly (produced by Batik) : it dumps identical image data
# multiple times into the svg. This directory helps to reduce them to the necessary
# memory comsumption
_imageData = {}

class Image(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - Image"
	def preserveAspectRatio(self,s) :
		self.props["keep_aspect"] = s
	def xlink__href(self,s) :
		#print s
		if s[:8] == "file:///" :
			self.props["uri"] = s.encode("UTF-8")
		elif s[:22] == "data:image/png;base64," :
			if s[22:] in _imageData :
				self.props["uri"] = _imageData[s[22:]] # use file reference
			else :
				# an ugly temporary file name, on windoze in %TEMP%
				fname = os.tempnam(None, "diapy-") + ".png"
				dd = s[22:].decode ("base64")
				f = open(fname, "wb")
				f.write(dd)
				f.close()
				# not really an uri but the reader appears to be robust enough ;-)
				_imageData[s[22:]] = "file:///" + fname
		else :
			pass #FIXME how to import data into dia ??
	def Create(self) :
		if not ("uri" in self.props or "data" in self.props) :
			return None
		return Object.Create(self)
	def ApplyProps(self,o) :
		if "width" in self.props :
			o.properties["elem_width"] = self.props["width"]
		if "width" in self.props :
			o.properties["elem_height"] = self.props["height"]
		if "uri" in self.props :
			o.properties["image_file"] = self.props["uri"][8:]
class Line(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - Line"
		# "line_width". "line_color"
		# "start_point". "end_point"
	def x1(self, s) :
		self.props["x"] = Scaled(s)
	def y1(self, s) :
		self.props["y"] = Scaled(s)
	def x2(self, s) :
		self.props["x2"] = Scaled(s)
	def y2(self, s) :
		self.props["y2"] = Scaled(s)
	def ApplyProps(self, o) :
		#pass
		o.properties["end_point"] = (self.props["x2"], self.props["y2"])
class Path(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - BezierLine" # or Beziergon ?
		self.pts = []
	def d(self, s) :
		self.props["data"] = s
		#FIXME: parse more - e.g. AQT - of the strange path data
		spd = rPathWhat.split(s)
		spw = rPathData.split(s)
		i = 1
		# current point
		xc = 0.0; yc = 0.0 # the current or second control point - ugly svg states ;(
		for s1 in spw :
			k = 0 # range further adjusted for last possibly empty -k-1
			if s1 == "M" : # moveto
				sp = rPathValue.split(spd[i])
				if sp[0] == "" : k = 1
				xc = Scaled(sp[k]); yc = Scaled(sp[k+1])
				self.pts.append((0, xc, yc))
			elif s1 == "L" : #lineto
				sp = rPathValue.split(spd[i])
				if sp[0] == "" : k = 1
				for j in range(k, len(sp)-k-1, 2) :
					xc = Scaled(sp[j]); yc = Scaled(sp[j+1])
					self.pts.append((1, xc, yc))
			elif s1 == "C" : # curveto
				sp = rPathValue.split(spd[i])
				if sp[0] == "" : k = 1
				for j in range(k, len(sp)-k-1, 6) :
					self.pts.append((2, Scaled(sp[j]), Scaled(sp[j+1]),
									Scaled(sp[j+2]), Scaled(sp[j+3]),
									Scaled(sp[j+4]), Scaled(sp[j+5])))
					# reflexion second control to current point, really ?
					xc =2 * Scaled(sp[j+4]) - Scaled(sp[j+2])
					yc =2 * Scaled(sp[j+5]) - Scaled(sp[j+3])
			elif s1 == "S" : # smooth curveto
				sp = rPathValue.split(spd[i])
				if sp[0] == "" : k = 1
				for j in range(k, len(sp)-k-1, 4) :
					x = Scaled(sp[j+2])
					y = Scaled(sp[j+3])
					x1 = Scaled(sp[j])
					y1 = Scaled(sp[j+1])
					self.pts.append((2, xc, yc,  # FIXME: current point ?
									x1, y1,
									x, y))
					xc = 2 * x - x1; yc = 2 * y - y1
			elif s1 == "z" or s1 == "Z" : # close
				self.dt = "Standard - Beziergon"
			elif s1 == "" : # too much whitespaces ;-)
				pass
			else :
				print("Huh?", s1)
				break
			i += 1
	def ApplyProps(self,o) :
		o.properties["bez_points"] = self.pts
	def Dump(self, indent) :
		print(" " * indent, self)
		for t in self.pts :
			print(" " * indent, t)
	#def Create(self) :
	#	return None # not yet
class Rect(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - Box"
		# "corner_radius",
	def ApplyProps(self,o) :
		o.properties["elem_width"] = self.props["width"]
		o.properties["elem_height"] = self.props["height"]
class Ellipse(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - Ellipse"
		self.props["cx"] = 0
		self.props["cy"] = 0
		self.props["rx"] = 1
		self.props["ry"] = 1
	def cx(self,s) :
		self.props["cx"] = Scaled(s)
		self.props["x"] = self.props["cx"] - self.props["rx"]
	def cy(self,s) :
		self.props["cy"] = Scaled(s)
		self.props["y"] = self.props["cy"] - self.props["ry"]
	def rx(self,s) :
		self.props["rx"] = Scaled(s)
		self.props["x"] = self.props["cx"] - self.props["rx"]
	def ry(self,s) :
		self.props["ry"] = Scaled(s)
		self.props["y"] = self.props["cy"] - self.props["ry"]
	def ApplyProps(self,o) :
		o.properties["elem_width"] = 2.0 * self.props["rx"]
		o.properties["elem_height"] = 2.0 * self.props["ry"]
class Circle(Ellipse) :
	def __init__(self) :
		Ellipse.__init__(self)
	def r(self,s) :
		Ellipse.rx(self,s)
		Ellipse.ry(self,s)
class Poly(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = None # abstract class !
	def points(self,s) :
		sp1 = s.split()
		pts = []
		for s1 in sp1 :
			sp2 = s1.split(",")
			if len(sp2) == 2 :
				pts.append((Scaled(sp2[0]), Scaled(sp2[1])))
		self.props["points"] = pts
	def ApplyProps(self,o) :
		o.properties["poly_points"] = self.props["points"]
class Polygon(Poly) :
	def __init__(self) :
		Poly.__init__(self)
		self.dt = "Standard - Polygon"
class Polyline(Poly) :
	def __init__(self) :
		Poly.__init__(self)
		self.dt = "Standard - PolyLine"
class Text(Object) :
	def __init__(self) :
		Object.__init__(self)
		self.dt = "Standard - Text"
		self.props["font-size"] = 1.0
		# text_font, text_height, text_color, text_alignment
	def Set(self, d) :
		if "text" in self.props :
			self.props["text"] += d
		else :
			self.props["text"] = d
	def text_anchor(self,s) :
		self.props["text-anchor"] = s
	def font_size(self,s) :
		global dfFontSize
		# ugh, just maintain another global state
		if s[-2:-1] != "e" : # FIXME ???
			dfFontSize = Scaled(s)
			#print "FontSize is", dfFontSize
		self.props["font-size"] = Scaled(s)
		# ?? self.props["y"] = self.props["y"] - Scaled(s)
	def font_weight(self, s) :
		self.props["font-weight"] = s
	def font_style(self, s) :
		self.props["font-style"] = s
	def font_family(self, s) :
		self.props["font-family"] = s
	def ApplyProps(self, o) :
		o.properties["text"] = self.props["text"].encode("UTF-8")
		if "text-anchor" in self.props :
			if self.props["text-anchor"] == "middle" : o.properties["text_alignment"] = 1
			elif self.props["text-anchor"] == "end" : o.properties["text_alignment"] = 2
			else : o.properties["text_alignment"] = 0
		if "fill" in self.props :
			o.properties["text_colour"] = Color(self.props["fill"])
		if "font-size" in self.props :
			o.properties["text_height"] = self.props["font-size"]
class Desc(Object) :
	#FIXME is this useful ?
	def __init__(self) :
		Object.__init__(self)
		self.dt = "UML - Note"
	def Set(self, d) :
		if "text" in self.props :
			self.props["text"] += d
		else :
			self.props["text"] = d
	def Create(self) :
		if "text" in self.props :
			pass
			#dia.message(0, self.props["text"].encode("UTF-8"))
		return None
class Title(Object) :
	#FIXME is this useful ?
	def __init__(self) :
		Object.__init__(self)
		self.dt = "UML - LargePackage"
	def Set(self, d) :
		if "text" in self.props :
			self.props["text"] += d
		else :
			self.props["text"] = d
	def Create(self) :
		if "text" in self.props :
			pass
		return None
class Unknown(Object) :
	def __init__(self, name) :
		Object.__init__(self)
		self.dt = "svg:" + name
	def Create(self) :
		return None

class Importer :
	def __init__(self) :
		self.errors = {}
		self.objects = []
	def Parse(self, sData) :
		import xml.parsers.expat
		ctx = []
		stack = []
		# 3 handler functions
		def start_element(name, attrs) :
			#print "<" + name + ">"
			if 0 == name.find("svg:") :
				name = name[4:]
			if len(stack) > 0 :
				grp = stack[-1]
			else :
				grp = None
			if 'g' == name :
				o = Group()
				stack.append(o)
			elif 'tspan' == name :
				#FIXME: to take all the style coming with it into account
				# Dia would need to support layouted text ...
				txn, txo = ctx[-1]
				if "dy" in attrs :
					txo.Set("" + "\n") # just a new line (best we can do?)
				elif "dx" in attrs :
					txo.Set(" ")
				ctx.append((txn, txo)) #push the same object
				return
			else :
				s = name.capitalize() + "()"
				try :
					# should be safe to use eval() here, by XML rules it can just be a name or would give
					# xml.parsers.expat.ExpatError: not well-formed (invalid token)
					o = eval(s)
				except :
					o = Unknown(name)
			if grp :
				grp.CopyProps(o)
			for a in attrs :
				if a == "class" : # eeek : keyword !
					st = cssStyle.Lookup(attrs[a])
					o.style(st)
					o.props[a] = attrs[a]
					continue
				ma = a.replace("-", "_")
				# e.g. xlink:href -> xlink__href
				ma = ma.replace(":", "__")
				s = "o." +  ma + "(\"" + attrs[a] + "\")"
				try :
					_eval(s, locals())
				except AttributeError as msg :
					o.props["meta"] = { a : attrs[a] }
					if msg not in self.errors :
						self.errors[msg] = s
				except SyntaxError as msg :
					if msg not in self.errors :
						self.errors[msg] = s
			if grp is None :
				self.objects.append(o)
			else :
				grp.Add(o)
			ctx.append((name, o)) #push
		def end_element(name) :
			if 'g' == name :
				del stack[-1]
			del ctx[-1] # pop
		def char_data(data):
			# may be called multiple times for one string
			ctx[-1][1].Set(data)

		p = xml.parsers.expat.ParserCreate()
		p.StartElementHandler = start_element
		p.EndElementHandler = end_element
		p.CharacterDataHandler = char_data

		p.Parse(sData)

	def Render(self,data) :
		layer = data.active_layer
		for o in self.objects :
			try :
				od = o.Create()
			except TypeError as e :
				od = None
				dia.message(1, "SVG import limited, consider another importer.\n(Error: " + str(e) + ")")
			if od :
				if o.translation :
					pos = od.properties["obj_pos"].value
					#FIXME:  looking at scascale.py this isn't completely correct
					x1 = pos.x + o.translation[0]
					y1 = pos.y + o.translation[1]
					od.move(x1, y1)
				layer.add_object(od)
		# create an 'Unhandled' layer and dump our Unknown
		# create an 'Errors' layer and dump our errors
		if len(list(self.errors.keys())) > 0 :
			layer = data.add_layer("Errors")
			s = "To hide the error messages delete or disable the 'Errors' layer\n"
			for e in list(self.errors.keys()) :
				s = s + str(e) + " -> " + str(self.errors[e]) + "\n"

			o = Text()
			o.props["fill"] = "red"
			o.Set(s)
			layer.add_object(o.Create())
		# create a 'Description' layer
		data.update_extents ()
		return 1
	def Dump(self) :
		for o in self.objects :
			o.Dump(0)
		for e in list(self.errors.keys()) :
			print(e, "->", self.errors[e])

def Test() :
	import sys
	imp = Importer()
	sName = sys.argv[1]
	if sName[-1] == "z" :
		import gzip
		f = gzip.open(sName)
	else :
		f = open(sName)
	imp.Parse(f.read())
	if len(sys.argv) > 2 :
		sys.stdout = open(sys.argv[2], "wb")
	imp.Dump()
	sys.exit(0)

if __name__ == '__main__': Test()

def import_svg(sFile, diagramData) :
	imp = Importer()
	f = open(sFile)
	imp.Parse(f.read())
	return imp.Render(diagramData)

def import_svgz(sFile, diagramData) :
	import gzip
	imp = Importer()
	f = gzip.open(sFile)
	imp.Parse(f.read())
	return imp.Render(diagramData)

import dia
dia.register_import(_("SVG plain"), "svg", import_svg)
dia.register_import(_("SVG compressed"), "svgz", import_svgz)




# A little debug helper marking and numbering the connection points
# of selected objects
# Copyright (c) 2013  Hans Breuer <hans@breuer.org>

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import dia

import gettext
_ = gettext.gettext

def mark_cps (data, flags) :
	objs = data.get_sorted_selected()
	layer = data.active_layer
	if len(objs) == 0 :
		dia.message (1, "Select objects for marking it's connection points!")
		return
	textType = dia.get_object_type("Standard - Text")
	for o in objs:
		for i in range(0, len(o.connections)):
			cp = o.connections[i]
			t, h1, h2 = textType.create (cp.pos.x, cp.pos.y)
			t.properties["text"] = str(i)
			# align the text based on the given directions
			if cp.flags & 0x3 : #CP_FLAGS_MAIN
				t.properties["text_vert_alignment"] = 3 # first line
			if cp.directions & 0x1: # north
				t.properties["text_vert_alignment"] = 1 # bottom
			elif cp.directions & 0x4: # south
				t.properties["text_vert_alignment"] = 0 # top
			else :
				t.properties["text_vert_alignment"] = 2 # center
			if cp.flags & 0x3 : #CP_FLAGS_MAIN
				t.properties["text_alignment"] = 1 # middle
			elif cp.directions & 0x2: # east
				t.properties["text_alignment"] = 0 # left
			elif cp.directions & 0x8: # west
				t.properties["text_alignment"] = 2 # right
			else :
				t.properties["text_alignment"] = 1 # middle
			# tint it with the connection point color
			if cp.flags & 0x3 : #CP_FLAGS_MAIN
				t.properties["text_colour"] = "red"
			elif cp.directions == 0 : # not necessarily a bug
				t.properties["text_colour"] = "green"
			else :
				t.properties["text_colour"] = "blue"
			# add it to the diagram
			layer.add_object(t)
			# connect the object with the cp at hand
			h2.connect(cp)
		# update the object and it's connected?
	data.update_extents ()
	adisp = dia.active_display()
	if adisp :
		adisp.diagram.update_extents()
		adisp.diagram.flush()


dia.register_action("DebugMarkConnectionPoints", _("_Mark Connection Points"),
		    "/DisplayMenu/Debug/DebugExtensionStart", mark_cps)



#    PyDia Self Documentation Series - Part I : PyDia itself
#    Copyright (c) 2003, 2005 , 2009 Hans Breuer  <hans@breuer.org>
#
#        generates a new diagram which contains all objects
#    of dir(dia). Now fills attributes and operations by using
#    Python reflexion ...
#

#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, math, dia, types

import gettext
_ = gettext.gettext

def distribute_objects (objs) :
	width = 0.0
	height = 0.0
	for o in objs :
		if width < o.properties["elem_width"].value :
			width = o.properties["elem_width"].value
		if height < o.properties["elem_height"].value :
			height = o.properties["elem_height"].value
	# add 20 % 'distance'
	width *= 1.2
	height *= 1.2
	area = len (objs) * width * height
	max_width = math.sqrt (area)
	x = 0.0
	y = 0.0
	dy = 0.0 # used to pack small objects more tightly
	for o in objs :
		if dy + o.properties["elem_height"].value * 1.2 > height :
			x += width
			dy = 0.0
		if x > max_width :
			x = 0.0
			y += height
		o.move (x, y + dy)
		dy += (o.properties["elem_height"].value * 1.2)
		if dy > .75 * height :
			x += width
			dy = 0.0
		if x > max_width :
			x = 0.0
			y += height

def find_class (data, name) :
	"Find an UML-Class with the given class-name"
	if not data :
		return None
	for la in data.layers :
		for o in la.objects :
			if o.type.name == 'UML - Class' and o.properties['name'].value == name :
				return o
	return None

def autodoc_fresh_cb (data, flags) :
	return autodoc_cb(data, flags, 0)
def autodoc_update_cb (data, flags) :
	return autodoc_cb(data, flags, data != None)

def autodoc_cb (data, flags, update) :
	show_comments = 1 # maybe this should be selectable
	if not data : # not set when called by the toolbox menu
		diagram = dia.new("PyDiaObjects.dia")
		# passed in data is not necessary valid - we are called from the Toolbox menu
		data = diagram
		display = diagram.display()
	else :
		diagram = None
		display = None
	layer = data.active_layer

	oType = dia.get_object_type ("UML - Class")

	theDir = dir(dia)
	# for reflection we need some objects ...
	theObjects = [data, layer, oType]
	try :
		theObjects.append (data.paper)
	except AttributeError :
		pass # no reason to fail with new bindings
	if diagram :
		theObjects.append (diagram)
	if display :
		theObjects.append (display)
	# add some objects with interesting properties
	#theObjects.append(dia.DiaImage())
	once = 1
	for s in ["Standard - Image", "Standard - BezierLine", "Standard - Text",
		"UML - Class", "UML - Dependency"] :
		o, h1, h2 = dia.get_object_type(s).create(0,0)
		for p in list(o.properties.keys()) :
			v = o.properties[p].value
			theObjects.append(v)
			if type(v) is tuple and len(v) > 0 :
				theObjects.append(v[0])
		if once :
			theObjects.append(o)
			theObjects.append(h1)
			theObjects.append(o.bounding_box)
			theObjects.append(o.connections[0])
			theObjects.append(o.handles[0])
			theObjects.append(o.properties)
			theObjects.append(o.properties["obj_pos"])
			once = 0
		# o is leaked here
	#print theObjects
	theTypes = {}
	for s in theDir :
		if s == "_dia" :
			continue # avoid all the messy details ;)
		if s in theTypes :
			continue
		for o in theObjects :
			is_a = eval("type(o) is dia." + s)
			#print s, o
			if is_a :
				theTypes[s] = o
				break
		if s not in theTypes :
			theTypes[s] = eval ("dia." + s)
	# add UML classes for every object in dir
	#print theTypes

	fresh = [] # list of newly created objects
	theGlobals = []
	# remove all objects prefixed with '__'
	for s in theDir :
		if s[:2] == '__' or s[:6] == '_swig_' :
			continue
		if s == "__doc__" or s == "__name__" :
			continue # hidden in the diagram objects but used below
		doc = eval("dia." + s + ".__doc__")
		is_a = eval("type(dia." + s + ") is types.BuiltinMethodType")
		if is_a :
			theGlobals.append((s,doc))
			continue
		# search the class we would create in the diagram and if found, modify rather than create
		o = None
		if update :
			o = find_class(data, s)
			h1 = h2 = None
			if o :
				show_comments = o.properties["visible_comments"].value
		if not o :
			o, h1, h2 = oType.create (0,0) # p.x, p.y
			layer.add_object (o)
			fresh.append(o)
		if doc :
			o.properties["comment"] = doc
		# set the objects name
		o.properties["name"] = s
		# now populate the object with ...
		if s in theTypes :
			t = theTypes[s]
			# ... methods and ...
			methods = []
			# ... attributes
			attributes = []
			members = dir(t)
			stereotype = ""
			if "__getitem__" in members :
				if "__len__" in members :
					stereotype = "sequence"
				elif "has_key" in members and "keys" in members :
					stereotype = "dictionary"
			for m in members :
				# again ignoring underscore prefixed
				if m[:2] == "__" or m == "thisown" : # ignore swig boilerplate
					continue
				try :
					is_m = eval("callable(t." + m + ")")
				except :
					print("type(t." + m + ")?")
					is_m = 0
				doc = ""
				tt = ""
				if 0 : # does not work well enough, giving only sometimes 'int' and often 'property'?
					try : # to detect the (return) type
						if is_m :
							oo = t()
							tt = eval("oo." + m + "().__class__.__name__")
						else :
							tt = eval("t." + m + ".__class__.__name__")
					except TypeError as msg :
						print(m, msg)
					except AttributeError as msg :
						print(m, msg) # No constructor defined
				try :
					# try to get the member's docstring from the type
					doc = eval("dia." + s + "." + m + ".__doc__")
					# fallback to object?
					#doc = eval("t." + m + ".__doc__")
				except AttributeError :
					doc = str(t) + "." + m
				if is_m : # (name, type, comment, stereotype, visibility, inheritance_type, query,class_scope, params)
					methods.append((m,tt,doc,'',0,0,0,0,()))
				else : # (name,type,value,comment,visibility,abstract,class_scope)
					attributes.append((m,tt,'',doc,0,0,0))
			o.properties["operations"] = methods
			o.properties["attributes"] = attributes
			o.properties["comment_line_length"] = 120
			o.properties["visible_comments"] = show_comments
			if stereotype != "" :
				o.properties["stereotype"] = stereotype
	o = None
	if update :
		o = find_class(data, 'dia')
	if o :
		show_comments = o.properties["visible_comments"].value
	else :
		# build the module object
		o, h1, h2 = oType.create (0,0) # p.x, p.y
		layer.add_object (o)
		# set the objects name
		o.properties["name"] = "dia"
		fresh.append(o)
	o.properties["comment"] = eval("dia.__doc__")
	o.properties["comment_line_length"] = 120
	o.properties["visible_comments"] = show_comments
	methods = []
	for s in theGlobals :
		if s[0].find("swigregister") >= 0 :
			continue # just noise
		methods.append((s[0],'',s[1],'',0,0,0,1,()))
	o.properties["operations"] = methods
	# all objects got there bounding box, distribute the fresh ones
	distribute_objects (fresh)

	if diagram :
		diagram.update_extents()
		diagram.flush()
	# work with bindings test
	return data

def autodoc_html_cb (data, flags) :
	import pydoc
	import os
	import webbrowser
	try :
		path = os.environ["TEMP"]
	except KeyError :
		# with os.tmpnam() we get a RuntimeWarning so fall back to
		path = "/tmp/"
	os.chdir(path)
	pydoc.writedoc(dia)
	dia.message(0, path + os.path.sep + "dia.html saved.")
	webbrowser.open('file://' + os.path.realpath(path + os.path.sep + "dia.html"))

dia.register_action ("HelpPydia2", _("PyDia _HTML Docs"),
                       "/ToolboxMenu/Help/HelpExtensionStart",
                       autodoc_html_cb)
dia.register_action ("HelpPydia", _("PyDia _Docs"),
                       "/ToolboxMenu/Help/HelpExtensionStart",
                       autodoc_fresh_cb)
dia.register_action ("UpdatePydia", _("PyDia Docs _Update"),
                       "/DisplayMenu/Help/HelpExtensionStart",
                       autodoc_update_cb)



# PyDia Simple Scale
# Copyright (c) 2003, Hans Breuer <hans@breuer.org>
#
# Experimental scaleing (selected) Objects via property api
#
# Known Issues:
#  - HANDLE_NON_MOVEABLE ?
#  - bezier control points
#  - unsizeable objects (or sizeable via multiple text size changing, e.g. UML Class)
#

#  This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import warnings
import dia

import gettext
_ = gettext.gettext

class CScaleDialog :
	def __init__(self, data) :
		import gi

		gi.require_version('Gtk', '3.0')

		with warnings.catch_warnings():
			warnings.filterwarnings("ignore", category=RuntimeWarning)
			from gi.repository import Gtk

		win = Gtk.Window()
		win.connect("delete_event", self.on_delete)
		win.set_title(_("Simple Scaling"))

		self.diagram = data
		self.win = win

		box1 = Gtk.VBox()
		win.add(box1)
		box1.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, True, True, 0)
		box2.show()

		self.entry = Gtk.Entry()
		self.entry.set_text("0.1")
		box2.pack_start(self.entry, True, True, 0)
		self.entry.show()

		separator = Gtk.HSeparator()
		box1.pack_start(separator, 0, True, 0)
		separator.show()

		box2 = Gtk.VBox(spacing=10)
		box2.set_border_width(10)
		box1.pack_start(box2, 0, True, 0)
		box2.show()

		button = Gtk.Button(_("Scale"))
		button.connect("clicked", self.on_scale)
		box2.pack_start(button, True, True, 0)
		button.set_can_default(True)
		button.grab_default()
		button.show()
		win.show()

	def on_scale(self, *args) :
		s = self.entry.get_text()
		scale = float(s)
		if scale > 0.001 and scale < 1000 :
			SimpleScale (self.diagram, float(s))
			self.diagram.update_extents ()
			self.diagram.flush()
		else :
			dia.message(1, "Value out of range!")
		self.win.destroy ()

	def on_delete (self, *args) :
		self.win.destroy ()

def ScaleLens(o, factor) :
	if "line_width" in o.properties :
		o.properties["line_width"] = o.properties["line_width"].value * factor
	if "text_height" in o.properties :
		o.properties["text_height"] = o.properties["text_height"].value * factor

def SimpleScale(data, factor) :
	objs = data.get_sorted_selected()
	if len(objs) == 0 :
		objs = data.active_layer.objects
	scaleFailed = {}
	for o in objs :
		pos = o.properties["obj_pos"].value
		hSE = None # the 'south east' handle to size the object
		if "elem_width" in o.properties :
			hLR = o.handles[7] # HANDLE_RESIZE_SE
			try :
				#if 0 == hLR.type : # HANDLE_NON_MOVABLE
				#	raise RuntimeError, "non moveable handle"
				x = pos.x * factor
				y = pos.y * factor
				x2 = x + (hLR.pos.x - pos.x) * factor
				y2 = y + (hLR.pos.y - pos.y) * factor
				# calculate all points before movement, handle is still connected and move moves it too
				o.move(x, y)
				o.move_handle(hLR, (x2, y2), 0, 0)
				ScaleLens(o, factor)
			except RuntimeError as msg :
				if o.type.name in scaleFailed :
					scaleFailed[o.type.name] += 1
				else :
					scaleFailed[o.type.name] = 1
		else :
			# must move all handles
			try :
				x = pos.x * factor
				y = pos.y * factor
				handles = []
				for h in o.handles :
					#if 0 == h.type : # HANDLE_NON_MOVABLE
					#	continue
					handles.append((h, (x + (h.pos.x - pos.x) * factor,
									  y + (h.pos.y - pos.y) * factor)))
				# handles are not necessary independent
				for h in handles :
					o.move_handle(h[0], h[1],  0, 0)
				ScaleLens(o, factor)
			except RuntimeError as msg :
				if o.type.name in scaleFailed :
					scaleFailed[o.type.name] += 1
				else :
					scaleFailed[o.type.name] = 1
	if len(list(scaleFailed.keys())) > 0 :
		sMsg = "Scaling failed for : "
		for s in list(scaleFailed.keys()) :
			sMsg = sMsg + "\n%s (%d)" % (s, scaleFailed[s])
		dia.message(1, sMsg)
	data.update_extents ()
	dia.active_display().add_update_all()

def scale_cb(data, flags) :
	dlg = CScaleDialog(data)

dia.register_action ("ObjectsSimplescaling", _("Simple _Scaling"),
		     "/DisplayMenu/Objects/ObjectsExtensionStart",
		     scale_cb)



#   Copyright (c) 2006, Hans Breuer <hans@breuer.org>
#
#   Draws bounding boxes of the objects in the active layer
#   into a new layer. Helps analyzing the font size problems
#
#    This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 2 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software
#   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import sys, dia, string

import gettext
_ = gettext.gettext

def bbox_cb (data, flags) :

	layer = data.active_layer
	dest = data.add_layer ("BBox of '%s' (%s)" % (layer.name, sys.platform), -1)
	box_type = dia.get_object_type ("Standard - Box")

	for o in layer.objects :
		bb = o.bounding_box
		b, h1, h2 = box_type.create (bb.left, bb.top)
		b.move_handle (b.handles[7], (bb.right, bb.bottom), 0, 0)
		b.properties["show_background"] = 0
		b.properties["line_width"] = 0
		b.properties["line_colour"] = 'red'
		dest.add_object (b)

def annotate_cb (data, flags) :

	layer = data.active_layer
	dest = data.add_layer ("Annotated '%s' (%s)" % (layer.name, sys.platform), -1)
	ann_type = dia.get_object_type ("Standard - Text")

	for o in layer.objects :
		bb = o.bounding_box
		a, h1, h2 = ann_type.create (bb.right, bb.top)

		a.properties["text"] = "h: %g w: %g" % (bb.bottom - bb.top, bb.right - bb.left)

		dest.add_object (a)

dia.register_action ("DrawBoundingbox", "_Draw BoundingBox",
                     "/DisplayMenu/Debug/DebugExtensionStart",
                     bbox_cb)

dia.register_action ("AnnotateMeasurements", "_Annotate",
                     "/DisplayMenu/Debug/DebugExtensionStart",
                     annotate_cb)
